import dataclasses
import re
import warnings
from abc import ABC
from collections import Counter, defaultdict
from collections.abc import Collection, Mapping
from datetime import date, datetime
from typing import ClassVar, Literal, TypeVar, cast, overload

from cognite.client.data_classes import data_modeling as dms
from cognite.client.data_classes.data_modeling import DataModelId, DataModelIdentifier, ViewId

from cognite.neat._client import NeatClient
from cognite.neat._client.data_classes.data_modeling import ContainerApplyDict, ViewApplyDict
from cognite.neat._constants import (
    COGNITE_MODELS,
    DMS_CONTAINER_PROPERTY_SIZE_LIMIT,
)
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._issues.warnings import NeatValueWarning
from cognite.neat._issues.warnings._models import (
    EnterpriseModelNotBuildOnTopOfCDMWarning,
    SolutionModelBuildOnTopOfCDMWarning,
)
from cognite.neat._rules._shared import (
    ReadInputRules,
    ReadRules,
    T_InputRules,
    VerifiedRules,
)
from cognite.neat._rules.analysis import DMSAnalysis
from cognite.neat._rules.importers import DMSImporter
from cognite.neat._rules.models import (
    DMSInputRules,
    DMSRules,
    InformationRules,
    SheetList,
    data_types,
)
from cognite.neat._rules.models.data_types import AnyURI, DataType, String
from cognite.neat._rules.models.dms import DMSMetadata, DMSProperty, DMSValidation, DMSView
from cognite.neat._rules.models.dms._rules import DMSContainer
from cognite.neat._rules.models.entities import (
    ClassEntity,
    ContainerEntity,
    DMSUnknownEntity,
    EdgeEntity,
    Entity,
    HasDataFilter,
    MultiValueTypeInfo,
    ReverseConnectionEntity,
    T_Entity,
    UnknownEntity,
    ViewEntity,
)
from cognite.neat._rules.models.information import InformationClass, InformationMetadata, InformationProperty
from cognite.neat._rules.models.information._rules_input import (
    InformationInputClass,
    InformationInputProperty,
    InformationInputRules,
)
from cognite.neat._utils.text import to_camel

from ._base import RulesTransformer
from ._verification import VerifyDMSRules

T_VerifiedInRules = TypeVar("T_VerifiedInRules", bound=VerifiedRules)
T_VerifiedOutRules = TypeVar("T_VerifiedOutRules", bound=VerifiedRules)
T_InputInRules = TypeVar("T_InputInRules", bound=ReadInputRules)
T_InputOutRules = TypeVar("T_InputOutRules", bound=ReadInputRules)


class ConversionTransformer(RulesTransformer[T_VerifiedInRules, T_VerifiedOutRules], ABC):
    """Base class for all conversion transformers."""

    ...


class ToCompliantEntities(RulesTransformer[ReadRules[InformationInputRules], ReadRules[InformationInputRules]]):  # type: ignore[misc]
    """Converts input rules to rules with compliant entity IDs that match regex patters used
    by DMS schema components."""

    @property
    def description(self) -> str:
        return "Ensures externalIDs are compliant with CDF"

    def transform(self, rules: ReadRules[InformationInputRules]) -> ReadRules[InformationInputRules]:
        if rules.rules is None:
            return rules
        copy: InformationInputRules = dataclasses.replace(rules.rules)
        copy.classes = self._fix_classes(copy.classes)
        copy.properties = self._fix_properties(copy.properties)
        return ReadRules(copy, rules.read_context)

    @classmethod
    def _fix_entity(cls, entity: str) -> str:
        entity = re.sub(r"[^_a-zA-Z0-9]+", "_", entity)

        # entity id must start with a letter
        if not entity[0].isalpha():
            entity = "prefix_" + entity
        # and end with a letter or number
        if not entity[-1].isalnum():
            entity = entity + "_suffix"

        # removing any double underscores that could occur
        return re.sub(r"[^a-zA-Z0-9]+", "_", entity)

    @classmethod
    def _fix_class(cls, class_: str | ClassEntity) -> str | ClassEntity:
        if isinstance(class_, str):
            if len(class_.split(":")) == 2:
                prefix, suffix = class_.split(":")
                class_ = f"{cls._fix_entity(prefix)}:{cls._fix_entity(suffix)}"

            else:
                class_ = cls._fix_entity(class_)

        elif isinstance(class_, ClassEntity) and type(class_.prefix) is str:
            class_ = ClassEntity(
                prefix=cls._fix_entity(class_.prefix),
                suffix=cls._fix_entity(class_.suffix),
            )

        return class_

    @classmethod
    def _fix_value_type(
        cls, value_type: str | DataType | ClassEntity | MultiValueTypeInfo
    ) -> str | DataType | ClassEntity | MultiValueTypeInfo:
        fixed_value_type: str | DataType | ClassEntity | MultiValueTypeInfo

        if isinstance(value_type, str):
            # this is a multi value type but as string
            if " | " in value_type:
                value_types = value_type.split(" | ")
                fixed_value_type = " | ".join([cast(str, cls._fix_value_type(v)) for v in value_types])
            # this is value type specified with prefix:suffix string
            elif ":" in value_type:
                fixed_value_type = cls._fix_class(value_type)

            # this is value type specified as suffix only
            else:
                fixed_value_type = cls._fix_entity(value_type)

        # value type specified as instances of DataType, ClassEntity or MultiValueTypeInfo
        elif isinstance(value_type, MultiValueTypeInfo):
            fixed_value_type = MultiValueTypeInfo(
                types=[cast(DataType | ClassEntity, cls._fix_value_type(type_)) for type_ in value_type.types],
            )
        elif isinstance(value_type, ClassEntity):
            fixed_value_type = cls._fix_class(value_type)

        # this is a DataType instance but also we should default to original value
        else:
            fixed_value_type = value_type

        return fixed_value_type

    @classmethod
    def _fix_classes(cls, definitions: list[InformationInputClass]) -> list[InformationInputClass]:
        fixed_definitions = []
        for definition in definitions:
            definition.class_ = cls._fix_class(definition.class_)
            fixed_definitions.append(definition)
        return fixed_definitions

    @classmethod
    def _fix_properties(cls, definitions: list[InformationInputProperty]) -> list[InformationInputProperty]:
        fixed_definitions = []
        for definition in definitions:
            definition.class_ = cls._fix_class(definition.class_)
            definition.property_ = cls._fix_entity(definition.property_)
            definition.value_type = cls._fix_value_type(definition.value_type)
            fixed_definitions.append(definition)
        return fixed_definitions


class PrefixEntities(RulesTransformer[ReadRules[T_InputRules], ReadRules[T_InputRules]]):  # type: ignore[type-var]
    """Prefixes all entities with a given prefix."""

    def __init__(self, prefix: str) -> None:
        self._prefix = prefix

    @property
    def description(self) -> str:
        return f"Prefixes all views with {self._prefix!r}"

    def transform(self, rules: ReadRules[T_InputRules]) -> ReadRules[T_InputRules]:
        in_ = rules.rules
        if in_ is None:
            return rules
        copy: T_InputRules = dataclasses.replace(in_)
        if isinstance(copy, InformationInputRules):
            prefixed_by_class: dict[str, str] = {}
            for cls in copy.classes:
                prefixed = str(self._with_prefix(cls.class_))
                prefixed_by_class[str(cls.class_)] = prefixed
                cls.class_ = prefixed
            for prop in copy.properties:
                prop.class_ = self._with_prefix(prop.class_)
                if str(prop.value_type) in prefixed_by_class:
                    prop.value_type = prefixed_by_class[str(prop.value_type)]
            return ReadRules(copy, rules.read_context)  # type: ignore[arg-type]
        elif isinstance(copy, DMSInputRules):
            prefixed_by_view: dict[str, str] = {}
            for view in copy.views:
                prefixed = str(self._with_prefix(view.view))
                prefixed_by_view[str(view.view)] = prefixed
                view.view = prefixed
            for dms_prop in copy.properties:
                dms_prop.view = self._with_prefix(dms_prop.view)
                if str(dms_prop.value_type) in prefixed_by_view:
                    dms_prop.value_type = prefixed_by_view[str(dms_prop.value_type)]
            if copy.containers:
                for container in copy.containers:
                    container.container = self._with_prefix(container.container)
            return ReadRules(copy, rules.read_context)
        raise NeatValueError(f"Unsupported rules type: {type(copy)}")

    @overload
    def _with_prefix(self, raw: str) -> str: ...

    @overload
    def _with_prefix(self, raw: T_Entity) -> T_Entity: ...

    def _with_prefix(self, raw: str | T_Entity) -> str | T_Entity:
        is_entity_format = not isinstance(raw, str)
        entity = Entity.load(raw)
        output: ClassEntity | ViewEntity | ContainerEntity
        if isinstance(entity, ClassEntity):
            output = ClassEntity(prefix=entity.prefix, suffix=f"{self._prefix}{entity.suffix}", version=entity.version)
        elif isinstance(entity, ViewEntity):
            output = ViewEntity(
                space=entity.space, externalId=f"{self._prefix}{entity.external_id}", version=entity.version
            )
        elif isinstance(entity, ContainerEntity):
            output = ContainerEntity(space=entity.space, externalId=f"{self._prefix}{entity.external_id}")
        elif isinstance(entity, UnknownEntity | Entity):
            return f"{self._prefix}{raw}"
        else:
            raise NeatValueError(f"Unsupported entity type: {type(entity)}")
        if is_entity_format:
            return cast(T_Entity, output)
        return str(output)


class InformationToDMS(ConversionTransformer[InformationRules, DMSRules]):
    """Converts InformationRules to DMSRules."""

    def __init__(self, ignore_undefined_value_types: bool = False, mode: Literal["edge_properties"] | None = None):
        self.ignore_undefined_value_types = ignore_undefined_value_types
        self.mode = mode

    def transform(self, rules: InformationRules) -> DMSRules:
        return _InformationRulesConverter(rules).as_dms_rules(self.ignore_undefined_value_types, self.mode)


class DMSToInformation(ConversionTransformer[DMSRules, InformationRules]):
    """Converts DMSRules to InformationRules."""

    def transform(self, rules: DMSRules) -> InformationRules:
        return _DMSRulesConverter(rules).as_information_rules()


class ConvertToRules(ConversionTransformer[VerifiedRules, VerifiedRules]):
    """Converts any rules to any rules."""

    def __init__(self, out_cls: type[VerifiedRules]):
        self._out_cls = out_cls

    def transform(self, rules: VerifiedRules) -> VerifiedRules:
        if isinstance(rules, self._out_cls):
            return rules
        if isinstance(rules, InformationRules) and self._out_cls is DMSRules:
            return InformationToDMS().transform(rules)
        if isinstance(rules, DMSRules) and self._out_cls is InformationRules:
            return DMSToInformation().transform(rules)
        raise ValueError(f"Unsupported conversion from {type(rules)} to {self._out_cls}")


_T_Entity = TypeVar("_T_Entity", bound=ClassEntity | ViewEntity)


class SetIDDMSModel(RulesTransformer[DMSRules, DMSRules]):
    def __init__(self, new_id: DataModelId | tuple[str, str, str]):
        self.new_id = DataModelId.load(new_id)

    @property
    def description(self) -> str:
        return f"Sets the Data Model ID to {self.new_id.as_tuple()}"

    def transform(self, rules: DMSRules) -> DMSRules:
        if self.new_id.version is None:
            raise NeatValueError("Version is required when setting a new Data Model ID")
        dump = rules.dump()
        dump["metadata"]["space"] = self.new_id.space
        dump["metadata"]["external_id"] = self.new_id.external_id
        dump["metadata"]["version"] = self.new_id.version
        # Serialize and deserialize to set the new space and external_id
        # as the default values for the new model.
        return DMSRules.model_validate(DMSInputRules.load(dump).dump())


class ToExtensionModel(RulesTransformer[DMSRules, DMSRules], ABC):
    type_: ClassVar[str]

    def __init__(self, new_model_id: DataModelIdentifier, org_name: str, dummy_property: str | None = None) -> None:
        self.new_model_id = DataModelId.load(new_model_id)
        if not self.new_model_id.version:
            raise NeatValueError("Version is required for the new model.")
        self.org_name = org_name
        self.dummy_property = dummy_property

    @property
    def description(self) -> str:
        return f"Prepared data model {self.new_model_id} to be {self.type_.replace('_', ' ')} data model."

    def _create_new_views(
        self, rules: DMSRules
    ) -> tuple[SheetList[DMSView], SheetList[DMSContainer], SheetList[DMSProperty]]:
        """Creates new views for the new model.

        If the dummy property is provided, it will also create a new container for each view
        with a single property that is the dummy property.
        """
        new_views = SheetList[DMSView]()
        new_containers = SheetList[DMSContainer]()
        new_properties = SheetList[DMSProperty]()

        for definition in rules.views:
            view_entity = self._remove_cognite_affix(definition.view)
            view_entity.version = cast(str, self.new_model_id.version)
            view_entity.prefix = self.new_model_id.space

            new_views.append(
                DMSView(
                    view=view_entity,
                    implements=[definition.view],
                    in_model=True,
                    name=definition.name,
                )
            )
            if self.dummy_property is None:
                continue

            container_entity = ContainerEntity(space=view_entity.prefix, externalId=view_entity.external_id)

            container = DMSContainer(container=container_entity)

            prefix = to_camel(view_entity.suffix)
            property_ = DMSProperty(
                view=view_entity,
                view_property=f"{prefix}{self.dummy_property}",
                value_type=String(),
                nullable=True,
                immutable=False,
                is_list=False,
                container=container_entity,
                container_property=f"{prefix}{self.dummy_property}",
            )

            new_properties.append(property_)
            new_containers.append(container)

        return new_views, new_containers, new_properties

    def _remove_cognite_affix(self, entity: _T_Entity) -> _T_Entity:
        """This method removes `Cognite` affix from the entity."""
        new_suffix = entity.suffix.replace("Cognite", self.org_name or "")
        if isinstance(entity, ViewEntity):
            return ViewEntity(space=entity.space, externalId=new_suffix, version=entity.version)  # type: ignore[return-value]
        elif isinstance(entity, ClassEntity):
            return ClassEntity(prefix=entity.prefix, suffix=new_suffix, version=entity.version)  # type: ignore[return-value]
        raise ValueError(f"Unsupported entity type: {type(entity)}")


class ToEnterpriseModel(ToExtensionModel):
    type_: ClassVar[str] = "enterprise"

    def __init__(
        self,
        new_model_id: DataModelIdentifier,
        org_name: str = "My",
        dummy_property: str = "GUID",
        move_connections: bool = False,
    ):
        super().__init__(new_model_id, org_name, dummy_property)
        self.move_connections = move_connections

    def transform(self, rules: DMSRules) -> DMSRules:
        reference_model_id = rules.metadata.as_data_model_id()
        if reference_model_id not in COGNITE_MODELS:
            warnings.warn(
                EnterpriseModelNotBuildOnTopOfCDMWarning(reference_model_id=reference_model_id).as_message(),
                stacklevel=2,
            )

        return self._to_enterprise(rules)

    def _to_enterprise(self, reference_model: DMSRules) -> DMSRules:
        dump = reference_model.dump()

        # This will create reference model components in the enterprise model space
        enterprise_model = DMSRules.model_validate(DMSInputRules.load(dump).dump())

        # Post validation metadata update:
        enterprise_model.metadata.name = self.type_
        enterprise_model.metadata.name = f"{self.org_name} {self.type_} data model"
        enterprise_model.metadata.space = self.new_model_id.space
        enterprise_model.metadata.external_id = self.new_model_id.external_id
        enterprise_model.metadata.version = cast(str, self.new_model_id.version)

        # Here we are creating enterprise views with a single container with a dummy property
        # for each view
        enterprise_views, enterprise_containers, enterprise_properties = self._create_new_views(enterprise_model)

        # We keep the reference views, and adding new enterprise views...
        enterprise_model.views.extend(enterprise_views)

        if self.move_connections:
            # Move connections from reference model to new enterprise model
            enterprise_properties.extend(self._move_connections(enterprise_model))

        # ... however, we do not want to keep the reference containers and properties
        enterprise_model.containers = enterprise_containers
        enterprise_model.properties = enterprise_properties

        return enterprise_model

    @staticmethod
    def _move_connections(rules: DMSRules) -> SheetList[DMSProperty]:
        implements: dict[ViewEntity, list[ViewEntity]] = defaultdict(list)
        new_properties = SheetList[DMSProperty]()

        for view in rules.views:
            if view.view.space == rules.metadata.space and view.implements:
                for implemented_view in view.implements:
                    implements.setdefault(implemented_view, []).append(view.view)

        # currently only supporting single implementation of reference view in enterprise view
        # connections that do not have properties
        if all(len(v) == 1 for v in implements.values()):
            for prop_ in rules.properties:
                if (
                    prop_.view.space != rules.metadata.space
                    and prop_.connection
                    and isinstance(prop_.value_type, ViewEntity)
                    and implements.get(prop_.view)
                    and implements.get(prop_.value_type)
                ):
                    if isinstance(prop_.connection, EdgeEntity) and prop_.connection.properties:
                        continue
                    new_property = prop_.model_copy(deep=True)
                    new_property.view = implements[prop_.view][0]
                    new_property.value_type = implements[prop_.value_type][0]
                    new_properties.append(new_property)

        return new_properties


class ToSolutionModel(ToExtensionModel):
    """Creates a solution data model based on an existing data model.

    The solution data model will create a new view for each view in the existing model.

    Args:
        new_model_id: DataData model identifier for the new model.
        org_name: If the existing model is a Cognite Data Model, this will replace the "Cognite" affix.
        mode: The mode of the solution model. Either "read" or "write". A "write" model will create a new
            container for each view with a dummy property. Read mode will only inherit the view filter from the
            original model.
        dummy_property: Only applicable if mode='write'. The identifier of the dummy property in the newly created
            container.
        exclude_views_in_other_spaces: Whether to exclude views that are not in the same space as the existing model,
            when creating the solution model.
        filter_type: If mode="read", this is the type of filter to apply to the new views. The filter is used to
            ensure that the new views will return the same instance as the original views. The view filter is the
            simplest filter, but it has limitation in the fusion UI. The container filter is in essence a more
            verbose version of the view filter, and it has better support in the fusion UI. The default is "container".

    """

    type_: ClassVar[str] = "solution"

    def __init__(
        self,
        new_model_id: DataModelIdentifier,
        org_name: str = "My",
        mode: Literal["read", "write"] = "read",
        dummy_property: str | None = "GUID",
        exclude_views_in_other_spaces: bool = True,
        filter_type: Literal["container", "view"] = "container",
    ):
        super().__init__(new_model_id, org_name, dummy_property if mode == "write" else None)
        self.mode = mode
        self.exclude_views_in_other_spaces = exclude_views_in_other_spaces
        self.filter_type = filter_type

    def transform(self, rules: DMSRules) -> DMSRules:
        reference_model = rules
        reference_model_id = reference_model.metadata.as_data_model_id()

        # if model is solution then we need to get correct space for views and containers
        if self.mode not in ["read", "write"]:
            raise NeatValueError(f"Unsupported mode: {self.mode}")

        if reference_model_id in COGNITE_MODELS:
            warnings.warn(
                SolutionModelBuildOnTopOfCDMWarning(reference_model_id=reference_model_id),
                stacklevel=2,
            )

        return self._to_solution(reference_model)

    @staticmethod
    def _has_views_in_multiple_space(rules: DMSRules) -> bool:
        return any(view.view.space != rules.metadata.space for view in rules.views)

    def _to_solution(self, reference_rules: DMSRules) -> DMSRules:
        """For creation of solution data model / rules specifically for mapping over existing containers."""

        dump = reference_rules.dump(entities_exclude_defaults=True)

        # Prepare new model metadata prior validation
        # Since we dropped the defaults, all entities will update the space and version
        # to the new model space and version
        dump["metadata"]["name"] = f"{self.org_name} {self.type_} data model"
        dump["metadata"]["space"] = self.new_model_id.space
        dump["metadata"]["external_id"] = self.new_model_id.external_id
        dump["metadata"]["version"] = self.new_model_id.version

        solution_model = DMSRules.model_validate(DMSInputRules.load(dump).dump())

        # This is not desirable for the containers, so we manually fix that here.
        # It is easier to change the space for all entities and then revert the containers, than
        # to change the space for all entities except the containers.
        for prop in solution_model.properties:
            if prop.container and prop.container.space == self.new_model_id.space:
                # If the container is in the new model space, we want to map it to the reference model space
                # This is reverting the .dump() -> .load() above.
                prop.container = ContainerEntity(
                    space=reference_rules.metadata.space,
                    externalId=prop.container.suffix,
                )

        for view in solution_model.views:
            view.implements = None

        if self.exclude_views_in_other_spaces and self._has_views_in_multiple_space(reference_rules):
            solution_model.views = SheetList[DMSView](
                [view for view in solution_model.views if view.view.space == solution_model.metadata.space]
            )

        # Dropping containers coming from reference model
        solution_model.containers = None

        # If reference model on which we are mapping one of Cognite Data Models
        # since we want to affix these with the organization name
        if reference_rules.metadata.as_data_model_id() in COGNITE_MODELS:
            # Remove Cognite affix in view external_id / suffix.
            for prop in solution_model.properties:
                prop.view = self._remove_cognite_affix(prop.view)
                if isinstance(prop.value_type, ViewEntity):
                    prop.value_type = self._remove_cognite_affix(prop.value_type)
            for view in solution_model.views:
                view.view = self._remove_cognite_affix(view.view)
                if view.implements:
                    view.implements = [self._remove_cognite_affix(implemented) for implemented in view.implements]

        if self.mode == "write":
            _, new_containers, new_properties = self._create_new_views(solution_model)
            # Here we add ONLY dummy properties of the solution model and
            # corresponding solution model space containers to hold them
            solution_model.containers = new_containers
            solution_model.properties.extend(new_properties)
        elif self.mode == "read":
            # Inherit view filter from original model to ensure the same instances are returned
            # when querying the new view.
            ref_views_by_external_id = {
                view.view.external_id: view
                for view in reference_rules.views
                if view.view.space == reference_rules.metadata.space
            }
            ref_containers_by_ref_view = defaultdict(set)
            for prop in reference_rules.properties:
                if prop.container:
                    ref_containers_by_ref_view[prop.view].add(prop.container)
            for view in solution_model.views:
                if ref_view := ref_views_by_external_id.get(view.view.external_id):
                    if self.filter_type == "view":
                        view.filter_ = HasDataFilter(inner=[ref_view.view])
                    elif self.filter_type == "container" and (
                        ref_containers := ref_containers_by_ref_view.get(ref_view.view)
                    ):
                        # Sorting to ensure deterministic order
                        view.filter_ = HasDataFilter(inner=sorted(ref_containers))

        return solution_model


class ToDataProductModel(ToSolutionModel):
    type_: ClassVar[str] = "data_product"

    def __init__(
        self,
        new_model_id: DataModelIdentifier,
        org_name: str = "My",
        include: Literal["same-space", "all"] = "same-space",
    ):
        super().__init__(new_model_id, org_name, mode="read", dummy_property=None, exclude_views_in_other_spaces=False)
        self.include = include

    def transform(self, rules: DMSRules) -> DMSRules:
        # Copy to ensure immutability
        expanded = self._expand_properties(rules.model_copy(deep=True))
        if self.include == "same-space":
            expanded.views = SheetList[DMSView](
                [view for view in expanded.views if view.view.space == expanded.metadata.space]
            )
            used_view_entities = {view.view for view in expanded.views}
            expanded.properties = SheetList[DMSProperty](
                [
                    prop
                    for prop in expanded.properties
                    if prop.view.space == expanded.metadata.space
                    and (
                        (isinstance(prop.value_type, ViewEntity) and prop.value_type in used_view_entities)
                        or not isinstance(prop.value_type, ViewEntity)
                    )
                ]
            )

        return self._to_solution(expanded)

    @staticmethod
    def _expand_properties(rules: DMSRules) -> DMSRules:
        probe = DMSAnalysis(rules)
        ancestor_properties_by_view = probe.classes_with_properties(
            consider_inheritance=True, allow_different_namespace=True
        )
        property_ids_by_view = {
            view: {prop.view_property for prop in properties}
            for view, properties in probe.classes_with_properties(consider_inheritance=False).items()
        }
        for view, property_ids in property_ids_by_view.items():
            ancestor_properties = ancestor_properties_by_view.get(view, [])
            for prop in ancestor_properties:
                if isinstance(prop.connection, ReverseConnectionEntity):
                    # If you try to add a reverse direct relation of a parent, it will fail as the ValueType of the
                    # original property will point to the parent view, and not the child.
                    continue
                if prop.view_property not in property_ids:
                    rules.properties.append(prop)
                    property_ids.add(prop.view_property)
        return rules


class ReduceCogniteModel(RulesTransformer[DMSRules, DMSRules]):
    _ASSET_VIEW = ViewId("cdf_cdm", "CogniteAsset", "v1")
    _VIEW_BY_COLLECTION: Mapping[Literal["3D", "Annotation", "BaseViews"], frozenset[ViewId]] = {
        "3D": frozenset(
            {
                ViewId("cdf_cdm", "Cognite3DModel", "v1"),
                ViewId("cdf_cdm", "Cognite3DObject", "v1"),
                ViewId("cdf_cdm", "Cognite3DRevision", "v1"),
                ViewId("cdf_cdm", "Cognite3DTransformation", "v1"),
                ViewId("cdf_cdm", "Cognite360Image", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageAnnotation", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageCollection", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageModel", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageStation", "v1"),
                ViewId("cdf_cdm", "CogniteCADModel", "v1"),
                ViewId("cdf_cdm", "CogniteCADNode", "v1"),
                ViewId("cdf_cdm", "CogniteCADRevision", "v1"),
                ViewId("cdf_cdm", "CogniteCubeMap", "v1"),
                ViewId("cdf_cdm", "CognitePointCloudModel", "v1"),
                ViewId("cdf_cdm", "CognitePointCloudRevision", "v1"),
                ViewId("cdf_cdm", "CognitePointCloudVolume", "v1"),
            }
        ),
        "Annotation": frozenset(
            {
                ViewId("cdf_cdm", "CogniteAnnotation", "v1"),
                ViewId("cdf_cdm", "CogniteDiagramAnnotation", "v1"),
            }
        ),
        "BaseViews": frozenset(
            {
                ViewId("cdf_cdm", "CogniteDescribable", "v1"),
                ViewId("cdf_cdm", "CogniteSchedulable", "v1"),
                ViewId("cdf_cdm", "CogniteSourceable", "v1"),
                ViewId("cdf_cdm", "CogniteVisualizable", "v1"),
            }
        ),
    }

    def __init__(self, drop: Collection[Literal["3D", "Annotation", "BaseViews"] | str]):
        self.drop_collection = cast(
            list[Literal["3D", "Annotation", "BaseViews"]],
            [collection for collection in drop if collection in self._VIEW_BY_COLLECTION],
        )
        self.drop_external_ids = {external_id for external_id in drop if external_id not in self._VIEW_BY_COLLECTION}

    def transform(self, rules: DMSRules) -> DMSRules:
        verified = rules
        if verified.metadata.as_data_model_id() not in COGNITE_MODELS:
            raise NeatValueError(f"Can only reduce Cognite Data Models, not {verified.metadata.as_data_model_id()}")

        exclude_views = {view for collection in self.drop_collection for view in self._VIEW_BY_COLLECTION[collection]}
        exclude_views |= {view.view.as_id() for view in verified.views if view.view.suffix in self.drop_external_ids}
        new_model = verified.model_copy(deep=True)

        properties_by_view = DMSAnalysis(new_model).classes_with_properties(consider_inheritance=True)

        new_model.views = SheetList[DMSView](
            [view for view in new_model.views if view.view.as_id() not in exclude_views]
        )
        new_properties = SheetList[DMSProperty]()

        for view in new_model.views:
            for prop in properties_by_view[view.view]:
                if self._is_asset_3D_property(prop):
                    # We filter out the 3D property of asset
                    continue
                new_properties.append(prop)

        new_model.properties = new_properties

        return new_model

    def _is_asset_3D_property(self, prop: DMSProperty) -> bool:
        if "3D" not in self.drop_collection:
            return False
        return prop.view.as_id() == self._ASSET_VIEW and prop.view_property == "object3D"

    @property
    def description(self) -> str:
        return f"Removed {len(self.drop_external_ids) + len(self.drop_collection)} views from data model"


class IncludeReferenced(RulesTransformer[DMSRules, DMSRules]):
    def __init__(self, client: NeatClient, include_properties: bool = False) -> None:
        self._client = client
        self.include_properties = include_properties

    def transform(self, rules: DMSRules) -> DMSRules:
        dms_rules = rules
        view_ids, container_ids = DMSValidation(dms_rules).imported_views_and_containers_ids()
        if not (view_ids or container_ids):
            warnings.warn(
                NeatValueWarning(
                    f"Data model {dms_rules.metadata.as_data_model_id()} does not have any "
                    "referenced views or containers."
                    "that is not already included in the data model."
                ),
                stacklevel=2,
            )
            return dms_rules

        schema = self._client.schema.retrieve([v.as_id() for v in view_ids], [c.as_id() for c in container_ids])
        copy_ = dms_rules.model_copy(deep=True)
        # Sorting to ensure deterministic order
        schema.containers = ContainerApplyDict(sorted(schema.containers.items(), key=lambda x: x[0].as_tuple()))
        schema.views = ViewApplyDict(sorted(schema.views.items(), key=lambda x: x[0].as_tuple()))
        importer = DMSImporter(schema)

        imported = importer.to_rules()
        if imported.rules is None:
            raise NeatValueError("Could not import the referenced views and containers.")

        verified = VerifyDMSRules(validate=False).transform(imported)
        if copy_.containers is None:
            copy_.containers = verified.containers
        else:
            existing_containers = {c.container for c in copy_.containers}
            copy_.containers.extend([c for c in verified.containers or [] if c.container not in existing_containers])
        existing_views = {v.view for v in copy_.views}
        copy_.views.extend([v for v in verified.views if v.view not in existing_views])
        if self.include_properties:
            existing_properties = {(p.view, p.view_property) for p in copy_.properties}
            copy_.properties.extend(
                [p for p in verified.properties if (p.view, p.view_property) not in existing_properties]
            )

        return copy_

    @property
    def description(self) -> str:
        return "Included referenced views and containers in the data model."


class AddClassImplements(RulesTransformer[InformationRules, InformationRules]):
    def __init__(self, implements: str, suffix: str):
        self.implements = implements
        self.suffix = suffix

    def transform(self, rules: InformationRules) -> InformationRules:
        info_rules = rules
        output = info_rules.model_copy(deep=True)
        for class_ in output.classes:
            if class_.class_.suffix.endswith(self.suffix):
                class_.implements = [ClassEntity(prefix=class_.class_.prefix, suffix=self.implements)]
        return output

    @property
    def description(self) -> str:
        return f"Added implements property to classes with suffix {self.suffix}"


class _InformationRulesConverter:
    _edge_properties: ClassVar[frozenset[str]] = frozenset({"endNode", "end_node", "startNode", "start_node"})

    def __init__(self, information: InformationRules):
        self.rules = information
        self.property_count_by_container: dict[ContainerEntity, int] = defaultdict(int)

    def as_dms_rules(
        self, ignore_undefined_value_types: bool = False, mode: Literal["edge_properties"] | None = None
    ) -> "DMSRules":
        from cognite.neat._rules.models.dms._rules import (
            DMSContainer,
            DMSProperty,
            DMSRules,
            DMSView,
        )

        info_metadata = self.rules.metadata
        default_version = info_metadata.version
        default_space = self._to_space(info_metadata.prefix)
        dms_metadata = self._convert_metadata_to_dms(info_metadata)
        edge_classes: set[ClassEntity] = set()
        property_to_edge: dict[tuple[ClassEntity, str], ClassEntity] = {}
        end_node_by_edge: dict[ClassEntity, ClassEntity] = {}
        if mode == "edge_properties":
            edge_classes = {
                cls_.class_ for cls_ in self.rules.classes if cls_.implements and cls_.implements[0].suffix == "Edge"
            }
            property_to_edge = {
                (prop.class_, prop.property_): prop.value_type
                for prop in self.rules.properties
                if prop.value_type in edge_classes and isinstance(prop.value_type, ClassEntity)
            }
            end_node_by_edge = {
                prop.class_: prop.value_type
                for prop in self.rules.properties
                if prop.class_ in edge_classes
                and (prop.property_ == "endNode" or prop.property_ == "end_node")
                and isinstance(prop.value_type, ClassEntity)
            }

        properties_by_class: dict[ClassEntity, list[DMSProperty]] = defaultdict(list)
        referenced_containers: dict[ContainerEntity, Counter[ClassEntity]] = defaultdict(Counter)
        for prop in self.rules.properties:
            if ignore_undefined_value_types and isinstance(prop.value_type, UnknownEntity):
                continue
            if prop.class_ in edge_classes and prop.property_ in self._edge_properties:
                continue
            dms_property = self._as_dms_property(
                prop, default_space, default_version, edge_classes, property_to_edge, end_node_by_edge
            )
            properties_by_class[prop.class_].append(dms_property)
            if dms_property.container:
                referenced_containers[dms_property.container][prop.class_] += 1

        views: list[DMSView] = []

        for cls_ in self.rules.classes:
            dms_view = DMSView(
                name=cls_.name,
                view=cls_.class_.as_view_entity(default_space, default_version),
                description=cls_.description,
                implements=self._get_view_implements(cls_, info_metadata, mode),
            )

            dms_view.logical = cls_.neatId
            cls_.physical = dms_view.neatId

            views.append(dms_view)

        class_by_entity = {cls_.class_: cls_ for cls_ in self.rules.classes}

        existing_containers: set[ContainerEntity] = set()

        containers: list[DMSContainer] = []
        for container_entity, class_entities in referenced_containers.items():
            if container_entity in existing_containers:
                continue
            constrains = self._create_container_constraint(
                class_entities, default_space, class_by_entity, referenced_containers
            )
            most_used_class_entity = class_entities.most_common(1)[0][0]
            class_ = class_by_entity[most_used_class_entity]
            container = DMSContainer(
                container=container_entity,
                name=class_.name,
                description=class_.description,
                constraint=constrains or None,
            )
            containers.append(container)

        return DMSRules(
            metadata=dms_metadata,
            properties=SheetList[DMSProperty]([prop for prop_set in properties_by_class.values() for prop in prop_set]),
            views=SheetList[DMSView](views),
            containers=SheetList[DMSContainer](containers),
        )

    @staticmethod
    def _create_container_constraint(
        class_entities: Counter[ClassEntity],
        default_space: str,
        class_by_entity: dict[ClassEntity, InformationClass],
        referenced_containers: Collection[ContainerEntity],
    ) -> list[ContainerEntity]:
        constrains: list[ContainerEntity] = []
        for entity in class_entities:
            class_ = class_by_entity[entity]
            for parent in class_.implements or []:
                parent_entity = parent.as_container_entity(default_space)
                if parent_entity in referenced_containers:
                    constrains.append(parent_entity)
        return constrains

    @classmethod
    def _convert_metadata_to_dms(cls, metadata: InformationMetadata) -> "DMSMetadata":
        from cognite.neat._rules.models.dms._rules import (
            DMSMetadata,
        )

        dms_metadata = DMSMetadata(
            space=metadata.space,
            version=metadata.version,
            external_id=metadata.external_id,
            creator=metadata.creator,
            name=metadata.name,
            created=metadata.created,
            updated=metadata.updated,
        )

        dms_metadata.logical = metadata.identifier
        metadata.physical = dms_metadata.identifier

        return dms_metadata

    def _as_dms_property(
        self,
        info_property: InformationProperty,
        default_space: str,
        default_version: str,
        edge_classes: set[ClassEntity],
        property_to_edge: dict[tuple[ClassEntity, str], ClassEntity],
        end_node_by_edge: dict[ClassEntity, ClassEntity],
    ) -> "DMSProperty":
        from cognite.neat._rules.models.dms._rules import DMSProperty

        # returns property type, which can be ObjectProperty or DatatypeProperty
        value_type = self._get_value_type(
            info_property,
            default_space,
            default_version,
            edge_classes,
            end_node_by_edge,
        )

        connection = self._get_connection(info_property, value_type, property_to_edge, default_space, default_version)

        container: ContainerEntity | None = None
        container_property: str | None = None
        is_list: bool | None = info_property.is_list
        nullable: bool | None = not info_property.is_mandatory
        if isinstance(connection, EdgeEntity):
            nullable = None
        elif connection == "direct":
            nullable = True
            container, container_property = self._get_container(info_property, default_space)
        else:
            container, container_property = self._get_container(info_property, default_space)

        dms_property = DMSProperty(
            name=info_property.name,
            value_type=value_type,
            nullable=nullable,
            is_list=is_list,
            connection=connection,
            default=info_property.default,
            container=container,
            container_property=container_property,
            view=info_property.class_.as_view_entity(default_space, default_version),
            view_property=info_property.property_,
        )

        # linking
        dms_property.logical = info_property.neatId
        info_property.physical = dms_property.neatId

        return dms_property

    @staticmethod
    def _get_connection(
        prop: InformationProperty,
        value_type: DataType | ViewEntity | DMSUnknownEntity,
        property_to_edge: dict[tuple[ClassEntity, str], ClassEntity],
        default_space: str,
        default_version: str,
    ) -> Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None:
        if isinstance(value_type, ViewEntity) and (prop.class_, prop.property_) in property_to_edge:
            edge_properties = property_to_edge[(prop.class_, prop.property_)]
            return EdgeEntity(properties=edge_properties.as_view_entity(default_space, default_version))
        if isinstance(value_type, ViewEntity) and prop.is_list:
            return EdgeEntity()
        elif isinstance(value_type, ViewEntity):
            return "direct"
        # defaulting to direct connection
        elif isinstance(value_type, DMSUnknownEntity):
            return "direct"
        return None

    def _get_value_type(
        self,
        prop: InformationProperty,
        default_space: str,
        default_version: str,
        edge_classes: set[ClassEntity],
        end_node_by_edge: dict[ClassEntity, ClassEntity],
    ) -> DataType | ViewEntity | DMSUnknownEntity:
        if isinstance(prop.value_type, DataType):
            return prop.value_type

        # UnknownEntity should  resolve to DMSUnknownEntity
        # meaning end node type is unknown
        elif isinstance(prop.value_type, UnknownEntity):
            return DMSUnknownEntity()

        elif isinstance(prop.value_type, ClassEntity) and (prop.value_type in edge_classes):
            if prop.value_type in end_node_by_edge:
                return end_node_by_edge[prop.value_type].as_view_entity(default_space, default_version)
            warnings.warn(
                NeatValueWarning(
                    f"Edge class {prop.value_type} does not have 'endNode' property, defaulting to DMSUnknownEntity"
                ),
                stacklevel=2,
            )
            return DMSUnknownEntity()
        elif isinstance(prop.value_type, ClassEntity):
            return prop.value_type.as_view_entity(default_space, default_version)

        elif isinstance(prop.value_type, MultiValueTypeInfo):
            # Multi Object type should resolve to DMSUnknownEntity
            # meaning end node type is unknown
            if prop.value_type.is_multi_object_type():
                non_unknown = [type_ for type_ in prop.value_type.types if isinstance(type_, UnknownEntity)]
                if list(non_unknown) == 1:
                    #
                    return non_unknown[0].as_view_entity(default_space, default_version)
                return DMSUnknownEntity()

            # Multi Data type should resolve to a single data type, or it should
            elif prop.value_type.is_multi_data_type():
                return self.convert_multi_data_type(prop.value_type)

            # Mixed types default to string
            else:
                non_any_uri = [type_ for type_ in prop.value_type.types if type_ != AnyURI()]
                if list(non_any_uri) == 1:
                    if isinstance(non_any_uri[0], ClassEntity):
                        return non_any_uri[0].as_view_entity(default_space, default_version)
                    else:
                        return non_any_uri[0]
                return String()

        raise ValueError(f"Unsupported value type: {prop.value_type.type_}")

    @classmethod
    def _to_space(cls, prefix: str) -> str:
        """Ensures that the prefix comply with the CDF space regex"""
        prefix = re.sub(r"[^a-zA-Z0-9_-]", "_", prefix)
        if prefix[0].isdigit() or prefix[0] == "_":
            prefix = f"a{prefix}"
        prefix = prefix[:43]
        if prefix[-1] == "_":
            prefix = f"{prefix[:-1]}1"
        return prefix

    def _get_container(self, prop: InformationProperty, default_space: str) -> tuple[ContainerEntity, str]:
        container_entity = prop.class_.as_container_entity(default_space)

        while self.property_count_by_container[container_entity] >= DMS_CONTAINER_PROPERTY_SIZE_LIMIT:
            container_entity.suffix = self._bump_suffix(container_entity.suffix)

        self.property_count_by_container[container_entity] += 1
        return container_entity, prop.property_

    def _get_view_implements(
        self, cls_: InformationClass, metadata: InformationMetadata, mode: Literal["edge_properties"] | None
    ) -> list[ViewEntity]:
        implements = []
        for parent in cls_.implements or []:
            if mode == "edge_properties" and parent.suffix == "Edge":
                continue
            view_entity = parent.as_view_entity(metadata.prefix, metadata.version)
            implements.append(view_entity)
        return implements

    @staticmethod
    def _bump_suffix(suffix: str) -> str:
        suffix_number = re.search(r"\d+$", suffix)

        if suffix_number:
            return suffix[: suffix_number.start()] + str(int(suffix_number.group()) + 1)
        else:
            return f"{suffix}2"

    @staticmethod
    def convert_multi_data_type(value_type: MultiValueTypeInfo) -> DataType:
        if not value_type.is_multi_data_type():
            raise ValueError("Only MultiValueType with DataType types is supported")
        # We check above that there are no ClassEntity types in the MultiValueType
        py_types = {type_.python for type_ in value_type.types}  # type: ignore[union-attr]

        # JSON mixed with other types should resolve to string that is safe choice
        if dms.Json in py_types and len(py_types) > 1:
            return data_types.String()
        elif dms.Json in py_types:
            return data_types.Json()
        elif not (py_types - {bool}):
            return data_types.Boolean()
        elif not (py_types - {int, bool}):
            return data_types.Integer()
        elif not (py_types - {float, int, bool}):
            return data_types.Double()
        elif not (py_types - {datetime, date}):
            return data_types.DateTime()

        return data_types.String()


class _DMSRulesConverter:
    def __init__(self, dms: DMSRules):
        self.dms = dms

    def as_information_rules(
        self,
    ) -> "InformationRules":
        from cognite.neat._rules.models.information._rules import (
            InformationClass,
            InformationProperty,
            InformationRules,
        )

        dms = self.dms.metadata

        metadata = self._convert_metadata_to_info(dms)

        classes = [
            InformationClass(
                # we do not want a version in class as we use URI for the class
                class_=ClassEntity(prefix=view.view.prefix, suffix=view.view.suffix),
                description=view.description,
                implements=[
                    # we do not want a version in class as we use URI for the class
                    implemented_view.as_class(skip_version=True)
                    for implemented_view in view.implements or []
                ],
            )
            for view in self.dms.views
        ]

        properties: list[InformationProperty] = []
        value_type: DataType | ClassEntity | str
        for property_ in self.dms.properties:
            if isinstance(property_.value_type, DataType):
                value_type = property_.value_type
            elif isinstance(property_.value_type, ViewEntity):
                value_type = ClassEntity(
                    prefix=property_.value_type.prefix,
                    suffix=property_.value_type.suffix,
                )
            elif isinstance(property_.value_type, DMSUnknownEntity):
                value_type = UnknownEntity()
            else:
                raise ValueError(f"Unsupported value type: {property_.value_type.type_}")

            properties.append(
                InformationProperty(
                    # Removing version
                    class_=ClassEntity(suffix=property_.view.suffix, prefix=property_.view.prefix),
                    property_=property_.view_property,
                    value_type=value_type,
                    description=property_.description,
                    min_count=(0 if property_.nullable or property_.nullable is None else 1),
                    max_count=(float("inf") if property_.is_list or property_.nullable is None else 1),
                )
            )

        return InformationRules(
            metadata=metadata,
            properties=SheetList[InformationProperty](properties),
            classes=SheetList[InformationClass](classes),
        )

    @classmethod
    def _convert_metadata_to_info(cls, metadata: DMSMetadata) -> "InformationMetadata":
        from cognite.neat._rules.models.information._rules import InformationMetadata

        return InformationMetadata(
            space=metadata.space,
            external_id=metadata.external_id,
            version=metadata.version,
            description=metadata.description,
            name=metadata.name,
            creator=metadata.creator,
            created=metadata.created,
            updated=metadata.updated,
        )



from rdflib import URIRef

from ._state import SessionState
from .exceptions import session_class_wrapper

try:
    from rich import print
except ImportError:
    ...


@session_class_wrapper
class DropAPI:
    """
    Drop instances from the session. Check out `.instances()` for performing the operation.
    """

    def __init__(self, state: SessionState):
        self._state = state

    def instances(self, type: str | list[str]) -> None:
        """Drop instances from the session.

        Args:
            type: The type of instances to drop.

        Example:
            ```python
            node_type_to_drop = "Pump"
            neat.drop.instances(node_type_to_drop)
            ```
        """
        type_list = type if isinstance(type, list) else [type]
        uri_type_type = dict((v, k) for k, v in self._state.instances.store.queries.types.items())
        selected_uri_by_type: dict[URIRef, str] = {}
        for type_item in type_list:
            if type_item not in uri_type_type:
                print(f"Type {type_item} not found.")
            selected_uri_by_type[uri_type_type[type_item]] = type_item

        result = self._state.instances.store.queries.drop_types(list(selected_uri_by_type.keys()))

        for type_uri, count in result.items():
            print(f"Dropped {count} instances of type {selected_uri_by_type[type_uri]}")
        return None



"""These are special exceptions that are used by the store to signal invalid transformers"""

from dataclasses import dataclass


class NeatStoreError(Exception):
    """Base class for all exceptions in the store module"""

    ...


@dataclass
class InvalidInputOperation(NeatStoreError, RuntimeError):
    """Raised when an invalid operation is attempted"""

    expected: tuple[type, ...]
    got: type


class EmptyStore(NeatStoreError, RuntimeError):
    """Raised when the store is empty"""

    ...



from collections import defaultdict
from collections.abc import Callable, Iterable, Set
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import Relationship, RelationshipList
from rdflib import Namespace

from cognite.neat._utils.auxiliary import create_sha256_hash

from ._base import DEFAULT_SKIP_METADATA_VALUES, ClassicCDFBaseExtractor, InstanceIdPrefix, T_CogniteResource


class RelationshipsExtractor(ClassicCDFBaseExtractor[Relationship]):
    """Extract data from Cognite Data Fusions Relationships into Neat."""

    _default_rdf_type = "Relationship"
    _instance_id_prefix = InstanceIdPrefix.relationship

    def __init__(
        self,
        items: Iterable[Relationship],
        namespace: Namespace | None = None,
        to_type: Callable[[Relationship], str | None] | None = None,
        total: int | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
    ):
        super().__init__(
            items,
            namespace=namespace,
            to_type=to_type,
            total=total,
            limit=limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
            camel_case=camel_case,
            as_write=as_write,
        )
        # This is used by the ClassicExtractor to log the target nodes, such
        # that it can extract them.
        # It is private to avoid exposing it to the user.
        self._log_target_nodes = False
        self._target_external_ids_by_type: dict[InstanceIdPrefix, set[str]] = defaultdict(set)

    @classmethod
    def _from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
    ) -> tuple[int | None, Iterable[Relationship]]:
        items = client.relationships(data_set_external_ids=data_set_external_id)
        return None, items

    @classmethod
    def _from_hierarchy(
        cls, client: CogniteClient, root_asset_external_id: str
    ) -> tuple[int | None, Iterable[T_CogniteResource]]:
        raise NotImplementedError("Relationships do not have a hierarchy.")

    @classmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[Relationship]]:
        relationships = RelationshipList.load(Path(file_path).read_text())
        return len(relationships), relationships

    def _fallback_id(self, item: Relationship) -> str | None:
        if item.external_id and item.source_external_id and item.target_external_id:
            if self._log_target_nodes and item.target_type and item.target_external_id:
                self._target_external_ids_by_type[InstanceIdPrefix.from_str(item.target_type)].add(
                    item.target_external_id
                )
            return create_sha256_hash(item.external_id)
        return None



from collections import defaultdict

from rdflib import URIRef

from cognite.neat._constants import DMS_LISTABLE_PROPERTY_LIMIT
from cognite.neat._rules.models.dms import DMSProperty, DMSRules, DMSView
from cognite.neat._rules.models.entities import ViewEntity

from ._base import BaseAnalysis


class DMSAnalysis(BaseAnalysis[DMSRules, DMSView, DMSProperty, ViewEntity, str]):
    """Assumes analysis over only the complete schema"""

    def _get_classes(self) -> list[DMSView]:
        return list(self.rules.views)

    def _get_properties(self) -> list[DMSProperty]:
        return list(self.rules.properties)

    def _get_cls_entity(self, class_: DMSView | DMSProperty) -> ViewEntity:
        return class_.view

    def _get_cls_parents(self, class_: DMSView) -> list[ViewEntity] | None:
        return list(class_.implements) if class_.implements else None

    @classmethod
    def _set_cls_entity(cls, property_: DMSProperty, class_: ViewEntity) -> None:
        property_.view = class_

    def _get_object(self, property_: DMSProperty) -> ViewEntity | None:
        return property_.value_type if isinstance(property_.value_type, ViewEntity) else None

    def _get_max_occurrence(self, property_: DMSProperty) -> int | float | None:
        return DMS_LISTABLE_PROPERTY_LIMIT if property_.is_list else 1

    def subset_rules(self, desired_classes: set[ViewEntity]) -> DMSRules:
        raise NotImplementedError()

    def _get_prop_entity(self, property_: DMSProperty) -> str:
        return property_.view_property

    def views_with_properties_linked_to_classes(
        self,
        consider_inheritance: bool = False,
        allow_different_namespace: bool = False,
    ) -> dict[ViewEntity, dict[str, URIRef]]:
        view_property_pairs = self.classes_with_properties(consider_inheritance, allow_different_namespace)

        view_and_properties_with_links: dict[ViewEntity, dict[str, URIRef]] = defaultdict(dict)

        for view, properties in view_property_pairs.items():
            view_and_properties_with_links[view] = {
                prop.view_property: prop.logical for prop in properties if prop.logical
            }

        return view_and_properties_with_links



from collections.abc import Collection
from pathlib import Path
from typing import Any, Literal, overload

from cognite.client.data_classes.data_modeling import SpaceApply

from cognite.neat._client import NeatClient
from cognite.neat._constants import COGNITE_MODELS
from cognite.neat._graph import loaders
from cognite.neat._rules import exporters
from cognite.neat._rules._constants import PATTERNS
from cognite.neat._rules._shared import VerifiedRules
from cognite.neat._rules.exporters._rules2dms import Component
from cognite.neat._rules.models.dms import DMSMetadata
from cognite.neat._utils.upload import UploadResultList

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class ToAPI:
    """API used to write the contents of a NeatSession to a specified destination. For instance writing information
    rules or DMS rules to a NEAT rules Excel spreadsheet, or writing a verified data model to CDF.

    """

    def __init__(self, state: SessionState, client: NeatClient | None, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self.cdf = CDFToAPI(state, client, verbose)

    def excel(
        self,
        io: Any,
        include_reference: bool = True,
    ) -> None:
        """Export the verified data model to Excel.

        Args:
            io: The file path or file-like object to write the Excel file to.
            include_reference: If True, the reference data model will be included. Defaults to True.
                Note that this only applies if you have created the data model using the
                .to_enterprise(), .to_solution(), or .to_data_product() methods.

        Example:
            Export information model to excel rules sheet
            ```python
            information_rules_file_name = "information_rules.xlsx"
            neat.to.excel(information_rules_file_name)
            ```

        Example:
            Read CogniteCore model, convert it to an enterprise model, and export it to an excel file
            ```python
            client = CogniteClient()
            neat = NeatSession(client)

            neat.read.cdf(("cdf_cdm", "CogniteCore", "v1"))
            neat.verify()
            neat.prepare.data_model.to_enterprise(
                data_model_id=("sp_doctrino_space", "ExtensionCore", "v1"),
                org_name="MyOrg",
                move_connections=True
            )
            dms_rules_file_name = "dms_rules.xlsx"
            neat.to.excel(dms_rules_file_name, include_reference=True)
            ```
        """
        reference_rules_with_prefix: tuple[VerifiedRules, str] | None = None
        if include_reference and self._state.last_reference:
            if (
                isinstance(self._state.last_reference.metadata, DMSMetadata)
                and self._state.last_reference.metadata.as_data_model_id() in COGNITE_MODELS
            ):
                prefix = "CDM"
            else:
                prefix = "Ref"
            reference_rules_with_prefix = self._state.last_reference, prefix

        exporter = exporters.ExcelExporter(styling="maximal", reference_rules_with_prefix=reference_rules_with_prefix)
        return self._state.rule_store.export_to_file(exporter, Path(io))

    @overload
    def yaml(self, io: None, format: Literal["neat"] = "neat", skip_system_spaces: bool = True) -> str: ...

    @overload
    def yaml(self, io: Any, format: Literal["neat", "toolkit"] = "neat", skip_system_spaces: bool = True) -> None: ...

    def yaml(
        self, io: Any | None = None, format: Literal["neat", "toolkit"] = "neat", skip_system_spaces: bool = True
    ) -> str | None:
        """Export the verified data model to YAML.

        Args:
            io: The file path or file-like object to write the YAML file to. Defaults to None.
            format: The format of the YAML file. Defaults to "neat".
            skip_system_spaces: If True, system spaces will be skipped. Defaults to True.

        !!! note "YAML formats"
            - "neat": This is the format Neat uses to store the data model.
            - "toolkit": This is the format used by Cognite Toolkit, that matches the CDF API.

        Returns:
            str | None: If io is None, the YAML string will be returned. Otherwise, None will be returned.

        Example:
            Export to yaml file in the case of "neat" format
            ```python
            your_yaml_file_name = "neat_rules.yaml"
            neat.to.yaml(your_yaml_file_name, format="neat")
            ```

        Example:
            Export yaml files as a zip folder in the case of "toolkit" format
            ```python
            your_zip_folder_name = "toolkit_data_model_files.zip"
            neat.to.yaml(your_zip_folder_name, format="toolkit")
            ```

        Example:
            Export yaml files to a folder in the case of "toolkit" format
            ```python
            your_folder_name = "my_project/data_model_files"
            neat.to.yaml(your_folder_name, format="toolkit")
            ```
        """
        if format == "neat":
            exporter = exporters.YAMLExporter()
            if io is None:
                return self._state.rule_store.export(exporter)

            self._state.rule_store.export_to_file(exporter, Path(io))
        elif format == "toolkit":
            if io is None or not isinstance(io, str | Path):
                raise NeatSessionError(
                    "Please provide a zip file or directory path to write the YAML files to."
                    "This is required for the 'toolkit' format."
                )
            user_path = Path(io)
            if user_path.suffix == "" and not user_path.exists():
                user_path.mkdir(parents=True)
            self._state.rule_store.export_to_file(
                exporters.DMSExporter(remove_cdf_spaces=skip_system_spaces), user_path
            )
        else:
            raise NeatSessionError("Please provide a valid format. 'neat' or 'toolkit'")

        return None


@session_class_wrapper
class CDFToAPI:
    """Write a verified Data Model and Instances to CDF."""

    def __init__(self, state: SessionState, client: NeatClient | None, verbose: bool) -> None:
        self._client = client
        self._state = state
        self._verbose = verbose

    def instances(self, space: str | None = None) -> UploadResultList:
        """Export the verified DMS instances to CDF.

        Args:
            space: Name of instance space to use. Default is to suffix the schema space with '_instances'.
            Note this space is required to be different than the space with the data model.

        """
        if not self._client:
            raise NeatSessionError("No CDF client provided!")

        space = space or f"{self._state.rule_store.last_verified_dms_rules.metadata.space}_instances"

        if space and space == self._state.rule_store.last_verified_dms_rules.metadata.space:
            raise NeatSessionError("Space for instances must be different from the data model space.")
        elif not PATTERNS.space_compliance.match(str(space)):
            raise NeatSessionError("Please provide a valid space name. {PATTERNS.space_compliance.pattern}")

        if not self._client.data_modeling.spaces.retrieve(space):
            self._client.data_modeling.spaces.apply(SpaceApply(space=space))

        loader = loaders.DMSLoader.from_rules(
            self._state.rule_store.last_verified_dms_rules,
            self._state.instances.store,
            instance_space=space,
            client=self._client,
        )
        result = loader.load_into_cdf(self._client)
        self._state.instances.outcome.append(result)
        print("You can inspect the details with the .inspect.outcome.instances(...) method.")
        return result

    def data_model(
        self,
        existing: Literal["fail", "skip", "update", "force", "recreate"] = "update",
        dry_run: bool = False,
        drop_data: bool = False,
        components: Component | Collection[Component] | None = None,
    ) -> UploadResultList:
        """Export the verified DMS data model to CDF.

        Args:
            existing: What to do if the component already exists. Defaults to "update".
                See the note below for more information about the options.
            dry_run: If True, no changes will be made to CDF. Defaults to False.
            drop_data: If existing is 'force' or 'recreate' and the operation will lead to data loss,
                the component will be skipped unless drop_data is True. Defaults to False.
                Note this only applies to spaces and containers if they contain data.
            components: The components to export. If None, all components will be exported. Defaults to None.

        !!! note "Data Model creation modes"
            - "fail": If any component already exists, the export will fail.
            - "skip": If any component already exists, it will be skipped.
            - "update": If any component already exists, it will be updated.
            - "force": If any component already exists, and the update fails, it will be deleted and recreated.
            - "recreate": All components will be deleted and recreated. The exception is spaces, which will be updated.

        """

        exporter = exporters.DMSExporter(existing=existing, export_components=components, drop_data=drop_data)

        if not self._client:
            raise NeatSessionError("No client provided!")

        result = self._state.rule_store.export_to_cdf(exporter, self._client, dry_run)
        print("You can inspect the details with the .inspect.outcome.data_model(...) method.")
        return result



import itertools
import json
import warnings
from collections import defaultdict
from collections.abc import Iterable, Sequence
from graphlib import TopologicalSorter
from pathlib import Path
from typing import Any, cast, get_args

import yaml
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.capabilities import Capability, DataModelInstancesAcl
from cognite.client.data_classes.data_modeling import ViewId
from cognite.client.data_classes.data_modeling.data_types import ListablePropertyType
from cognite.client.data_classes.data_modeling.ids import InstanceId
from cognite.client.data_classes.data_modeling.views import SingleEdgeConnection
from cognite.client.exceptions import CogniteAPIError
from pydantic import BaseModel, ValidationInfo, create_model, field_validator
from rdflib import RDF, URIRef

from cognite.neat._client import NeatClient
from cognite.neat._constants import DMS_DIRECT_RELATION_LIST_LIMIT, is_readonly_property
from cognite.neat._graph._tracking import LogTracker, Tracker
from cognite.neat._issues import IssueList, NeatIssue, NeatIssueList
from cognite.neat._issues.errors import (
    ResourceConversionError,
    ResourceCreationError,
    ResourceDuplicatedError,
    ResourceRetrievalError,
)
from cognite.neat._issues.warnings import PropertyDirectRelationLimitWarning, PropertyTypeNotSupportedWarning
from cognite.neat._rules.analysis._dms import DMSAnalysis
from cognite.neat._rules.models import DMSRules
from cognite.neat._rules.models.data_types import _DATA_TYPE_BY_DMS_TYPE, Json
from cognite.neat._rules.models.entities._single_value import ViewEntity
from cognite.neat._shared import InstanceType
from cognite.neat._store import NeatGraphStore
from cognite.neat._utils.auxiliary import create_sha256_hash
from cognite.neat._utils.rdf_ import remove_namespace_from_uri
from cognite.neat._utils.upload import UploadResult

from ._base import _END_OF_CLASS, CDFLoader


class DMSLoader(CDFLoader[dm.InstanceApply]):
    """Loads Instances to Cognite Data Fusion Data Model Service from NeatGraph.

    Args:
        graph_store (NeatGraphStore): The graph store to load the data into.
        data_model (dm.DataModel[dm.View] | None): The data model to load.
        instance_space (str): The instance space to load the data into.
        class_neat_id_by_view_id (dict[ViewId, URIRef] | None): A mapping from view id to class name. Defaults to None.
        create_issues (Sequence[NeatIssue] | None): A list of issues that occurred during reading. Defaults to None.
        tracker (type[Tracker] | None): The tracker to use. Defaults to None.
        rules (DMSRules | None): The DMS rules used by the data model. This is used to lookup the
            instances in the store. Defaults to None.
        client (NeatClient | None): This is used to lookup containers such that the loader
            creates instances in accordance with required constraints. Defaults to None.
    """

    def __init__(
        self,
        graph_store: NeatGraphStore,
        data_model: dm.DataModel[dm.View] | None,
        instance_space: str,
        class_neat_id_by_view_id: dict[ViewId, URIRef] | None = None,
        create_issues: Sequence[NeatIssue] | None = None,
        tracker: type[Tracker] | None = None,
        rules: DMSRules | None = None,
        client: NeatClient | None = None,
    ):
        super().__init__(graph_store)
        self.data_model = data_model
        self.instance_space = instance_space
        self.class_neat_id_by_view_id = class_neat_id_by_view_id or {}
        self._issues = IssueList(create_issues or [])
        self._tracker: type[Tracker] = tracker or LogTracker
        self.rules = rules
        self._client = client

    @classmethod
    def from_data_model_id(
        cls,
        client: NeatClient,
        data_model_id: dm.DataModelId,
        graph_store: NeatGraphStore,
        instance_space: str,
    ) -> "DMSLoader":
        issues: list[NeatIssue] = []
        data_model: dm.DataModel[dm.View] | None = None
        try:
            data_model = client.data_modeling.data_models.retrieve(data_model_id, inline_views=True).latest_version()
        except Exception as e:
            issues.append(ResourceRetrievalError(data_model_id, "data model", str(e)))

        return cls(graph_store, data_model, instance_space, {}, issues, client=client)

    @classmethod
    def from_rules(
        cls, rules: DMSRules, graph_store: NeatGraphStore, instance_space: str, client: NeatClient | None = None
    ) -> "DMSLoader":
        issues: list[NeatIssue] = []
        data_model: dm.DataModel[dm.View] | None = None
        try:
            data_model = rules.as_schema().as_read_model()
        except Exception as e:
            issues.append(
                ResourceConversionError(
                    identifier=rules.metadata.as_identifier(),
                    resource_type="DMS Rules",
                    target_format="read DMS model",
                    reason=str(e),
                )
            )

        class_neat_id_by_view_id = {view.view.as_id(): view.logical for view in rules.views if view.logical}

        return cls(
            graph_store,
            data_model,
            instance_space,
            class_neat_id_by_view_id,
            issues,
            rules=rules,
            client=client,
        )

    def _load(self, stop_on_exception: bool = False) -> Iterable[dm.InstanceApply | NeatIssue | type[_END_OF_CLASS]]:
        if self._issues.has_errors and stop_on_exception:
            raise self._issues.as_exception()
        elif self._issues.has_errors:
            yield from self._issues
            return
        if not self.data_model:
            # There should already be an error in this case.
            return

        views_with_linked_properties = (
            DMSAnalysis(self.rules).views_with_properties_linked_to_classes(consider_inheritance=True)
            if self.rules and self.rules.metadata.logical
            else None
        )
        view_and_count_by_id = self._select_views_with_instances(self.data_model.views)
        if self._client:
            view_and_count_by_id, properties_point_to_self = self._sort_by_direct_relation_dependencies(
                view_and_count_by_id
            )
        else:
            properties_point_to_self = {}

        view_ids: list[str] = []
        for view_id in view_and_count_by_id.keys():
            view_ids.append(repr(view_id))
            if view_id in properties_point_to_self:
                # If the views have a dependency on themselves, we need to run it twice.
                view_ids.append(f"{view_id!r} (self)")

        tracker = self._tracker(type(self).__name__, view_ids, "views")
        for view_id, (view, _) in view_and_count_by_id.items():
            pydantic_cls, edge_by_type, issues = self._create_validation_classes(view)  # type: ignore[var-annotated]
            yield from issues
            tracker.issue(issues)

            if view_id in properties_point_to_self:
                # If the view has a dependency on itself, we need to run it twice.
                # First, to ensure that all nodes are created, and then to add the direct relations.
                # This only applies if there is a require constraint on the container, if not
                # we can create an empty node on the fly.
                iterations = [properties_point_to_self[view_id], set()]
            else:
                iterations = [set()]

            for skip_properties in iterations:
                if skip_properties:
                    track_id = f"{view_id} (self)"
                else:
                    track_id = repr(view_id)
                tracker.start(track_id)
                if views_with_linked_properties:
                    # we need graceful exit if the view is not in the view_property_pairs
                    property_link_pairs = views_with_linked_properties.get(ViewEntity.from_id(view_id))

                    if class_neat_id := self.class_neat_id_by_view_id.get(view_id):
                        reader = self.graph_store._read_via_rules_linkage(class_neat_id, property_link_pairs)
                    else:
                        error_view = ResourceRetrievalError(view_id, "view", "View not linked to class")
                        tracker.issue(error_view)
                        if stop_on_exception:
                            raise error_view
                        yield error_view
                        continue
                else:
                    # this assumes no changes in the suffix of view and class
                    reader = self.graph_store.read(view.external_id)

                for identifier, properties in reader:
                    if skip_properties:
                        properties = {k: v for k, v in properties.items() if k not in skip_properties}
                    try:
                        yield self._create_node(identifier, properties, pydantic_cls, view_id)
                    except ValueError as e:
                        error_node = ResourceCreationError(identifier, "node", error=str(e))
                        tracker.issue(error_node)
                        if stop_on_exception:
                            raise error_node from e
                        yield error_node
                    yield from self._create_edges(identifier, properties, edge_by_type, tracker)
                tracker.finish(track_id)
                yield _END_OF_CLASS

    def write_to_file(self, filepath: Path) -> None:
        if filepath.suffix not in [".json", ".yaml", ".yml"]:
            raise ValueError(f"File format {filepath.suffix} is not supported")
        dumped: dict[str, list] = {"nodes": [], "edges": [], "issues": []}
        for item in self.load(stop_on_exception=False):
            key = {
                dm.NodeApply: "nodes",
                dm.EdgeApply: "edges",
                NeatIssue: "issues",
            }.get(type(item))
            if key is None:
                # This should never happen, and is a bug in neat
                raise ValueError(f"Item {item} is not supported. This is a bug in neat please report it.")
            dumped[key].append(item.dump())
        with filepath.open("w", encoding=self._encoding, newline=self._new_line) as f:
            if filepath.suffix == ".json":
                json.dump(dumped, f, indent=2)
            else:
                yaml.safe_dump(dumped, f, sort_keys=False)

    def _select_views_with_instances(self, views: list[dm.View]) -> dict[dm.ViewId, tuple[dm.View, int]]:
        """Selects the views with data."""
        view_and_count_by_id: dict[dm.ViewId, tuple[dm.View, int]] = {}
        uri_by_type: dict[str, URIRef] = {
            remove_namespace_from_uri(uri[0]): uri[0]  # type: ignore[misc]
            for uri in self.graph_store.queries.list_types()
        }
        for view in views:
            view_id = view.as_id()
            neat_id = self.class_neat_id_by_view_id.get(view_id)
            if neat_id is not None:
                count = self.graph_store.count_of_id(neat_id)
            elif view.external_id in uri_by_type:
                count = self.graph_store.count_of_type(uri_by_type[view.external_id])
            else:
                continue
            if count > 0:
                view_and_count_by_id[view_id] = view, count

        return view_and_count_by_id

    def _sort_by_direct_relation_dependencies(
        self, view_and_count_by_id: dict[dm.ViewId, tuple[dm.View, int]]
    ) -> tuple[dict[dm.ViewId, tuple[dm.View, int]], dict[dm.ViewId, set[str]]]:
        """Sorts the views by container constraints."""
        if not self._client:
            return view_and_count_by_id, {}
        # We need to retrieve the views to ensure we get all properties, such that we can find all
        # the containers that the view is linked to.
        views = self._client.data_modeling.views.retrieve(
            list(view_and_count_by_id.keys()), include_inherited_properties=True
        )
        container_ids_by_view_id = {view.as_id(): view.referenced_containers() for view in views}
        referenced_containers = {
            container for containers in container_ids_by_view_id.values() for container in containers
        }
        containers = self._client.data_modeling.containers.retrieve(list(referenced_containers))
        container_by_id = {container.as_id(): container for container in containers}

        dependency_on_self: dict[dm.ViewId, set[str]] = defaultdict(set)
        view_id_by_dependencies: dict[dm.ViewId, set[dm.ViewId]] = {}
        for view in views:
            view_id = view.as_id()
            dependencies = set()
            for prop_id, prop in view.properties.items():
                if isinstance(prop, dm.MappedProperty) and isinstance(prop.type, dm.DirectRelation) and prop.source:
                    container = container_by_id[prop.container]
                    has_require_constraint = any(
                        isinstance(constraint, dm.RequiresConstraint) for constraint in container.constraints.values()
                    )
                    if has_require_constraint and prop.source == view_id:
                        dependency_on_self[view_id].add(prop_id)
                    elif has_require_constraint:
                        dependencies.add(prop.source)
            view_id_by_dependencies[view_id] = dependencies

        ordered_view_ids = TopologicalSorter(view_id_by_dependencies).static_order()

        return {
            view_id: view_and_count_by_id[view_id] for view_id in ordered_view_ids if view_id in view_and_count_by_id
        }, dict(dependency_on_self)

    def _create_validation_classes(
        self, view: dm.View
    ) -> tuple[type[BaseModel], dict[str, tuple[str, dm.EdgeConnection]], NeatIssueList]:
        issues = IssueList()
        field_definitions: dict[str, tuple[type, Any]] = {}
        edge_by_property: dict[str, tuple[str, dm.EdgeConnection]] = {}
        validators: dict[str, classmethod] = {}
        direct_relation_by_property: dict[str, dm.DirectRelation] = {}
        unit_properties: list[str] = []
        json_fields: list[str] = []
        for prop_id, prop in view.properties.items():
            if isinstance(prop, dm.EdgeConnection):
                edge_by_property[prop_id] = prop_id, prop
            if isinstance(prop, dm.MappedProperty):
                if is_readonly_property(prop.container, prop.container_property_identifier):
                    continue

                if isinstance(prop.type, dm.DirectRelation):
                    if prop.container == dm.ContainerId("cdf_cdm", "CogniteTimeSeries") and prop_id == "unit":
                        unit_properties.append(prop_id)
                    else:
                        direct_relation_by_property[prop_id] = prop.type
                    python_type: Any = dict
                else:
                    data_type = _DATA_TYPE_BY_DMS_TYPE.get(prop.type._type)
                    if not data_type:
                        issues.append(
                            PropertyTypeNotSupportedWarning(
                                view.as_id(),
                                "view",
                                prop_id,
                                prop.type._type,
                            )
                        )
                        continue

                    if data_type == Json:
                        json_fields.append(prop_id)
                    python_type = data_type.python
                if isinstance(prop.type, ListablePropertyType) and prop.type.is_list:
                    python_type = list[python_type]
                default_value: Any = prop.default_value
                if prop.nullable:
                    python_type = python_type | None
                else:
                    default_value = ...

                field_definitions[prop_id] = (python_type, default_value)

        def parse_list(cls, value: Any, info: ValidationInfo) -> list[str]:
            if isinstance(value, list) and list.__name__ not in _get_field_value_types(cls, info):
                if len(value) == 1:
                    return value[0]
                raise ValueError(f"Got multiple values for {info.field_name}: {value}")

            return value

        def parse_json_string(cls, value: Any, info: ValidationInfo) -> dict | list:
            if isinstance(value, dict):
                return value
            elif isinstance(value, list):
                try:
                    return [json.loads(v) if isinstance(v, str) else v for v in value]
                except json.JSONDecodeError as error:
                    raise ValueError(f"Not valid JSON string for {info.field_name}: {value}, error {error}") from error
            elif isinstance(value, str):
                try:
                    return json.loads(value)
                except json.JSONDecodeError as error:
                    raise ValueError(f"Not valid JSON string for {info.field_name}: {value}, error {error}") from error
            else:
                raise ValueError(f"Expect valid JSON string or dict for {info.field_name}: {value}")

        if json_fields:
            validators["parse_json_string"] = field_validator(*json_fields, mode="before")(parse_json_string)  # type: ignore[assignment, arg-type]

        validators["parse_list"] = field_validator("*", mode="before")(parse_list)  # type: ignore[assignment, arg-type]

        if direct_relation_by_property:

            def parse_direct_relation(cls, value: list, info: ValidationInfo) -> dict | list[dict]:
                # We validate above that we only get one value for single direct relations.
                if list.__name__ in _get_field_value_types(cls, info):
                    result = [{"space": self.instance_space, "externalId": remove_namespace_from_uri(v)} for v in value]
                    if len(result) <= DMS_DIRECT_RELATION_LIST_LIMIT:
                        return result
                    warnings.warn(
                        PropertyDirectRelationLimitWarning(
                            identifier="unknown",
                            resource_type="view property",
                            property_name=cast(str, cls.model_fields[info.field_name].alias or info.field_name),
                            limit=DMS_DIRECT_RELATION_LIST_LIMIT,
                        ),
                        stacklevel=2,
                    )
                    # To get deterministic results, we sort by space and externalId
                    result.sort(key=lambda x: (x["space"], x["externalId"]))
                    return result[:DMS_DIRECT_RELATION_LIST_LIMIT]
                elif value:
                    return {"space": self.instance_space, "externalId": remove_namespace_from_uri(value[0])}
                return {}

            validators["parse_direct_relation"] = field_validator(*direct_relation_by_property.keys(), mode="before")(  # type: ignore[assignment]
                parse_direct_relation  # type: ignore[arg-type]
            )

        if unit_properties:

            def parse_direct_relation_to_unit(cls, value: Any, info: ValidationInfo) -> dict | list[dict]:
                if value:
                    return {"space": "cdf_cdm_units", "externalId": remove_namespace_from_uri(value[0])}
                return {}

            validators["parse_direct_relation_to_unit"] = field_validator(*unit_properties, mode="before")(  # type: ignore[assignment]
                parse_direct_relation_to_unit  # type: ignore[arg-type]
            )

        pydantic_cls = create_model(view.external_id, __validators__=validators, **field_definitions)  # type: ignore[arg-type, call-overload]
        return pydantic_cls, edge_by_property, issues

    def _create_node(
        self,
        identifier: str,
        properties: dict[str | InstanceType, list[str]],
        pydantic_cls: type[BaseModel],
        view_id: dm.ViewId,
    ) -> dm.InstanceApply:
        type_ = properties.pop(RDF.type, [None])[0]
        created = pydantic_cls.model_validate(properties)

        return dm.NodeApply(
            space=self.instance_space,
            external_id=identifier,
            type=(dm.DirectRelationReference(view_id.space, view_id.external_id) if type_ is not None else None),
            sources=[
                dm.NodeOrEdgeData(source=view_id, properties=dict(created.model_dump(exclude_unset=True).items()))
            ],
        )

    def _create_edges(
        self,
        identifier: str,
        properties: dict[str, list[str]],
        edge_by_type: dict[str, tuple[str, dm.EdgeConnection]],
        tracker: Tracker,
    ) -> Iterable[dm.EdgeApply | NeatIssue]:
        for predicate, values in properties.items():
            if predicate not in edge_by_type:
                continue
            prop_id, edge = edge_by_type[predicate]
            if isinstance(edge, SingleEdgeConnection) and len(values) > 1:
                error = ResourceDuplicatedError(
                    resource_type="edge",
                    identifier=identifier,
                    location=f"Multiple values for single edge {edge}. Expected only one.",
                )
                tracker.issue(error)
                yield error
            for target in values:
                external_id = f"{identifier}.{prop_id}.{target}"
                yield dm.EdgeApply(
                    space=self.instance_space,
                    external_id=(external_id if len(external_id) < 256 else create_sha256_hash(external_id)),
                    type=edge.type,
                    start_node=dm.DirectRelationReference(self.instance_space, identifier),
                    end_node=dm.DirectRelationReference(self.instance_space, target),
                )

    def _get_required_capabilities(self) -> list[Capability]:
        return [
            DataModelInstancesAcl(
                actions=[
                    DataModelInstancesAcl.Action.Write,
                    DataModelInstancesAcl.Action.Write_Properties,
                    DataModelInstancesAcl.Action.Read,
                ],
                scope=DataModelInstancesAcl.Scope.SpaceID([self.instance_space]),
            )
        ]

    def _upload_to_cdf(
        self,
        client: CogniteClient,
        items: list[dm.InstanceApply],
        dry_run: bool,
        read_issues: IssueList,
    ) -> Iterable[UploadResult]:
        nodes: list[dm.NodeApply] = []
        edges: list[dm.EdgeApply] = []
        source_by_node_id: dict[dm.NodeId, str] = {}
        source_by_edge_id: dict[dm.EdgeId, str] = {}
        for item in items:
            if isinstance(item, dm.NodeApply):
                nodes.append(item)
                if item.sources:
                    source_by_node_id[item.as_id()] = item.sources[0].source.external_id
                else:
                    source_by_node_id[item.as_id()] = "node"
            elif isinstance(item, dm.EdgeApply):
                edges.append(item)
                if item.sources:
                    source_by_edge_id[item.as_id()] = item.sources[0].source.external_id
                else:
                    source_by_edge_id[item.as_id()] = "edge"
        try:
            upserted = client.data_modeling.instances.apply(
                nodes,
                edges,
                auto_create_end_nodes=True,
                auto_create_start_nodes=True,
                skip_on_version_conflict=True,
            )
        except CogniteAPIError as e:
            result = UploadResult[InstanceId](name="Instances", issues=read_issues)
            result.error_messages.append(str(e))
            result.failed_upserted.update(item.as_id() for item in e.failed + e.unknown)
            result.created.update(item.as_id() for item in e.successful)
            yield result
        else:
            for instances, ids_by_source in [
                (upserted.nodes, source_by_node_id),
                (upserted.edges, source_by_edge_id),
            ]:
                for name, subinstances in itertools.groupby(
                    sorted(instances, key=lambda i: ids_by_source.get(i.as_id(), "")),  # type: ignore[call-overload, index, attr-defined]
                    key=lambda i: ids_by_source.get(i.as_id(), ""),  # type: ignore[index, attr-defined]
                ):
                    result = UploadResult(name=name, issues=read_issues)
                    for instance in subinstances:  # type: ignore[attr-defined]
                        if instance.was_modified and instance.created_time == instance.last_updated_time:
                            result.created.add(instance.as_id())
                        elif instance.was_modified:
                            result.changed.add(instance.as_id())
                        else:
                            result.unchanged.add(instance.as_id())
                    yield result


def _get_field_value_types(cls, info):
    return [type_.__name__ for type_ in get_args(cls.model_fields[info.field_name].annotation)]



import sys
import warnings
from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import cast

import pandas as pd
from pandas import Index
from rdflib import Dataset, Namespace, URIRef
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._graph._shared import rdflib_to_oxi_type
from cognite.neat._graph.extractors import RdfFileExtractor, TripleExtractors
from cognite.neat._graph.queries import Queries
from cognite.neat._graph.transformers import Transformers
from cognite.neat._rules.analysis import InformationAnalysis
from cognite.neat._rules.models import InformationRules
from cognite.neat._rules.models.entities import ClassEntity
from cognite.neat._shared import InstanceType, Triple
from cognite.neat._utils.auxiliary import local_import
from cognite.neat._utils.rdf_ import add_triples_in_batch

from ._provenance import Change, Provenance

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class NeatGraphStore:
    """NeatGraphStore is a class that stores the graph and provides methods to read/write data it contains


    Args:
        graph : Instance of rdflib.Graph class for graph storage
        rules:
    """

    rdf_store_type: str

    def __init__(
        self,
        graph: Dataset,
        rules: InformationRules | None = None,
    ):
        self.rules: InformationRules | None = None

        _start = datetime.now(timezone.utc)
        self.graph = graph
        self.provenance = Provenance(
            [
                Change.record(
                    activity=f"{type(self).__name__}.__init__",
                    start=_start,
                    end=datetime.now(timezone.utc),
                    description=f"Initialize graph store as {type(self.graph.store).__name__}",
                )
            ]
        )

        if rules:
            self.add_rules(rules)
        else:
            self.base_namespace = DEFAULT_NAMESPACE

        self.queries = Queries(self.graph, self.rules)

    @property
    def type_(self) -> str:
        "Return type of the graph store"
        return type(self.graph.store).__name__

    def add_rules(self, rules: InformationRules) -> None:
        """This method is used to add rules to the graph store and it is the only correct
        way to add rules to the graph store, after the graph store has been initialized.
        """

        self.rules = rules
        self.base_namespace = self.rules.metadata.namespace
        self.queries = Queries(self.graph, self.rules)
        self.provenance.append(
            Change.record(
                activity=f"{type(self)}.rules",
                start=datetime.now(timezone.utc),
                end=datetime.now(timezone.utc),
                description=f"Added rules to graph store as {type(self.rules).__name__}",
            )
        )

        if self.rules.prefixes:
            self._upsert_prefixes(self.rules.prefixes)

    def _upsert_prefixes(self, prefixes: dict[str, Namespace]) -> None:
        """Adds prefixes to the graph store."""
        _start = datetime.now(timezone.utc)
        for prefix, namespace in prefixes.items():
            self.graph.bind(prefix, namespace)

        self.provenance.append(
            Change.record(
                activity=f"{type(self).__name__}._upsert_prefixes",
                start=_start,
                end=datetime.now(timezone.utc),
                description="Upsert prefixes to graph store",
            )
        )

    @classmethod
    def from_memory_store(cls, rules: InformationRules | None = None) -> "Self":
        return cls(Dataset(), rules)

    @classmethod
    def from_sparql_store(
        cls,
        query_endpoint: str | None = None,
        update_endpoint: str | None = None,
        returnFormat: str = "csv",
        rules: InformationRules | None = None,
    ) -> "Self":
        store = SPARQLUpdateStore(
            query_endpoint=query_endpoint,
            update_endpoint=update_endpoint,
            returnFormat=returnFormat,
            context_aware=False,
            postAsEncoded=False,
            autocommit=False,
        )
        graph = Dataset(store=store)
        return cls(graph, rules)

    @classmethod
    def from_oxi_store(cls, storage_dir: Path | None = None, rules: InformationRules | None = None) -> "Self":
        """Creates a NeatGraphStore from an Oxigraph store."""
        local_import("pyoxigraph", "oxi")
        local_import("oxrdflib", "oxi")
        import oxrdflib
        import pyoxigraph

        # Adding support for both oxigraph in-memory and file-based storage
        for i in range(4):
            try:
                oxi_store = pyoxigraph.Store(path=str(storage_dir) if storage_dir else None)
                break
            except OSError as e:
                if "lock" in str(e) and i < 3:
                    continue
                raise e
        else:
            raise Exception("Error initializing Oxigraph store")

        graph = Dataset(
            store=oxrdflib.OxigraphStore(store=oxi_store),
        )

        return cls(graph, rules)

    def write(self, extractor: TripleExtractors) -> None:
        _start = datetime.now(timezone.utc)
        success = True

        if isinstance(extractor, RdfFileExtractor) and not extractor.issue_list.has_errors:
            self._parse_file(extractor.filepath, cast(str, extractor.format), extractor.base_uri)
        elif isinstance(extractor, RdfFileExtractor):
            success = False
            issue_text = "\n".join([issue.as_message() for issue in extractor.issue_list])
            warnings.warn(
                f"Cannot write to graph store with {type(extractor).__name__}, errors found in file:\n{issue_text}",
                stacklevel=2,
            )
        else:
            self._add_triples(extractor.extract())

        if success:
            _end = datetime.now(timezone.utc)
            # Need to do the hasattr in case the extractor comes from NeatEngine.
            activities = (
                extractor._get_activity_names()
                if hasattr(extractor, "_get_activity_names")
                else [type(extractor).__name__]
            )
            for activity in activities:
                self.provenance.append(
                    Change.record(
                        activity=activity,
                        start=_start,
                        end=_end,
                        description=f"Extracted triples to graph store using {type(extractor).__name__}",
                    )
                )

    def _read_via_rules_linkage(
        self, class_neat_id: URIRef, property_link_pairs: dict[str, URIRef] | None
    ) -> Iterable[tuple[str, dict[str | InstanceType, list[str]]]]:
        if self.rules is None:
            warnings.warn("Rules not found in graph store! Aborting!", stacklevel=2)
            return
        if self.multi_type_instances:
            warnings.warn(
                "Multi typed instances detected, issues with loading can occur!",
                stacklevel=2,
            )

        if cls := InformationAnalysis(self.rules).classes_by_neat_id.get(class_neat_id):
            if property_link_pairs:
                property_renaming_config = {
                    prop_uri: prop_name
                    for prop_name, prop_neat_id in property_link_pairs.items()
                    if (
                        prop_uri := InformationAnalysis(self.rules).neat_id_to_transformation_property_uri(prop_neat_id)
                    )
                }

                yield from self._read_via_class_entity(cls.class_, property_renaming_config)
                return
            else:
                warnings.warn("Rules not linked", stacklevel=2)
                return
        else:
            warnings.warn("Class with neat id {class_neat_id} found in rules", stacklevel=2)
            return

    def _read_via_class_entity(
        self,
        class_entity: ClassEntity,
        property_renaming_config: dict[URIRef, str] | None = None,
    ) -> Iterable[tuple[str, dict[str | InstanceType, list[str]]]]:
        if self.rules is None:
            warnings.warn("Rules not found in graph store!", stacklevel=2)
            return
        if self.multi_type_instances:
            warnings.warn(
                "Multi typed instances detected, issues with loading can occur!",
                stacklevel=2,
            )

        if class_entity not in [definition.class_ for definition in self.rules.classes]:
            warnings.warn("Desired type not found in graph!", stacklevel=2)
            return

        if not (class_uri := InformationAnalysis(self.rules).class_uri(class_entity)):
            warnings.warn(
                f"Class {class_entity.suffix} does not have namespace defined for prefix {class_entity.prefix} Rules!",
                stacklevel=2,
            )
            return

        has_hop_transformations = InformationAnalysis(self.rules).has_hop_transformations()
        has_self_reference_transformations = InformationAnalysis(
            self.rules
        ).has_self_reference_property_transformations()
        if has_hop_transformations or has_self_reference_transformations:
            msg = (
                f"Rules contain [{'Hop' if has_hop_transformations else '' }"
                f", {'SelfReferenceProperty' if has_self_reference_transformations else '' }]"
                " rdfpath."
                f" Run [{'ReduceHopTraversal' if has_hop_transformations else '' }"
                f", {'AddSelfReferenceProperty' if has_self_reference_transformations else '' }]"
                " transformer(s) first!"
            )

            warnings.warn(
                msg,
                stacklevel=2,
            )
            return

        # get all the instances for give class_uri
        instance_ids = self.queries.list_instances_ids_of_class(class_uri)

        # get potential property renaming config
        property_renaming_config = property_renaming_config or InformationAnalysis(
            self.rules
        ).define_property_renaming_config(class_entity)

        # get property types to guide process of removing or not namespaces from results
        property_types = InformationAnalysis(self.rules).property_types(class_entity)
        for instance_id in instance_ids:
            if res := self.queries.describe(
                instance_id=instance_id,
                instance_type=class_entity.suffix,
                property_renaming_config=property_renaming_config,
                property_types=property_types,
            ):
                yield res

    def read(
        self,
        class_: str,
    ) -> Iterable[tuple[str, dict[str | InstanceType, list[str]]]]:
        """Read instances for given class from the graph store.

        !!! note "Assumption"
            This method assumes that the class_ belongs to the same (name)space as
            the rules which are attached to the graph store.

        """

        if not self.rules:
            warnings.warn("Rules not found in graph store!", stacklevel=2)
            return
        if self.multi_type_instances:
            warnings.warn(
                "Multi typed instances detected, issues with loading can occur!",
                stacklevel=2,
            )

        class_entity = ClassEntity(prefix=self.rules.metadata.prefix, suffix=class_)

        if class_entity not in [definition.class_ for definition in self.rules.classes]:
            warnings.warn("Desired type not found in graph!", stacklevel=2)
            return

        yield from self._read_via_class_entity(class_entity)

    def count_of_id(self, neat_id: URIRef) -> int:
        """Count the number of instances of a given type

        Args:
            neat_id: Type for which instances are to be counted

        Returns:
            Number of instances
        """
        if not self.rules:
            warnings.warn("Rules not found in graph store!", stacklevel=2)
            return 0

        class_entity = next(
            (definition.class_ for definition in self.rules.classes if definition.neatId == neat_id), None
        )
        if not class_entity:
            warnings.warn("Desired type not found in graph!", stacklevel=2)
            return 0

        if not (class_uri := InformationAnalysis(self.rules).class_uri(class_entity)):
            warnings.warn(
                f"Class {class_entity.suffix} does not have namespace defined for prefix {class_entity.prefix} Rules!",
                stacklevel=2,
            )
            return 0

        return self.count_of_type(class_uri)

    def count_of_type(self, class_uri: URIRef) -> int:
        query = f"SELECT (COUNT(?instance) AS ?instanceCount) WHERE {{ ?instance a <{class_uri}> }}"
        return int(next(iter(self.graph.query(query)))[0])  # type: ignore[arg-type, index]

    def _parse_file(
        self,
        filepath: Path,
        format: str = "turtle",
        base_uri: URIRef | None = None,
    ) -> None:
        """Imports graph data from file.

        Args:
            filepath : File path to file containing graph data, by default None
            format : rdflib format file containing RDF graph, by default "turtle"
            base_uri : base URI to add to graph in case of relative URIs, by default None

        !!! note "Oxigraph store"
            By default we are using non-transactional mode for parsing RDF files.
            This gives us a significant performance boost when importing large RDF files.
            Underhood of rdflib we are triggering oxrdflib plugin which in respect
            calls `bulk_load` method from oxigraph store. See more at:
            https://pyoxigraph.readthedocs.io/en/stable/store.html#pyoxigraph.Store.bulk_load
        """

        # Oxigraph store, do not want to type hint this as it is an optional dependency
        if type(self.graph.store).__name__ == "OxigraphStore":
            local_import("pyoxigraph", "oxi")

            # this is necessary to trigger rdflib oxigraph plugin
            self.graph.parse(
                filepath,
                format=rdflib_to_oxi_type(format),
                transactional=False,
                publicID=base_uri,
            )
            self.graph.store._store.optimize()  # type: ignore[attr-defined]

        # All other stores
        else:
            if filepath.is_file():
                self.graph.parse(filepath, publicID=base_uri)
            else:
                for filename in filepath.iterdir():
                    if filename.is_file():
                        self.graph.parse(filename, publicID=base_uri)

    def _add_triples(self, triples: Iterable[Triple], batch_size: int = 10_000):
        """Adds triples to the graph store in batches.

        Args:
            triples: list of triples to be added to the graph store
            batch_size: Batch size of triples per commit, by default 10_000
            verbose: Verbose mode, by default False
        """
        add_triples_in_batch(self.graph, triples, batch_size)

    def transform(self, transformer: Transformers) -> None:
        """Transforms the graph store using a transformer."""

        missing_changes = [
            change for change in transformer._need_changes if not self.provenance.activity_took_place(change)
        ]
        if self.provenance.activity_took_place(type(transformer).__name__) and transformer._use_only_once:
            warnings.warn(
                f"Cannot transform graph store with {type(transformer).__name__}, already applied",
                stacklevel=2,
            )
        elif missing_changes:
            warnings.warn(
                (
                    f"Cannot transform graph store with {type(transformer).__name__}, "
                    f"missing one or more required changes [{', '.join(missing_changes)}]"
                ),
                stacklevel=2,
            )

        else:
            _start = datetime.now(timezone.utc)
            transformer.transform(self.graph)
            self.provenance.append(
                Change.record(
                    activity=f"{type(transformer).__name__}",
                    start=_start,
                    end=datetime.now(timezone.utc),
                    description=transformer.description,
                )
            )

    @property
    def summary(self) -> pd.DataFrame:
        return pd.DataFrame(self.queries.summarize_instances(), columns=["Type", "Occurrence"])

    @property
    def multi_type_instances(self) -> dict[str, list[str]]:
        return self.queries.multi_type_instances()

    def _repr_html_(self) -> str:
        provenance = self.provenance._repr_html_()
        summary: pd.DataFrame = self.summary

        if summary.empty:
            summary_text = "<br /><strong>Graph is empty</strong><br />"
        else:
            summary_text = (
                "<br /><strong>Overview</strong>:"  # type: ignore
                f"<ul><li>{len(summary)} types</strong></li>"
                f"<li>{sum(summary['Occurrence'])} instances</strong></li></ul>"
                f"{cast(pd.DataFrame, self._shorten_summary(summary))._repr_html_()}"  # type: ignore[operator]
            )

        if self.multi_type_instances:
            summary_text += "<br><strong>Multi value instances detected! Loading could have issues!</strong></br>"  # type: ignore

        return f"{summary_text}" f"{provenance}"

    def _shorten_summary(self, summary: pd.DataFrame) -> pd.DataFrame:
        """Shorten summary to top 5 types by occurrence."""
        top_5_rows = summary.head(5)

        indexes = [
            *top_5_rows.index.tolist(),
        ]
        data = [
            top_5_rows,
        ]
        if len(summary) > 6:
            last_row = summary.tail(1)
            indexes += [
                "...",
                *last_row.index.tolist(),
            ]
            data.extend([pd.DataFrame([["..."] * summary.shape[1]], columns=summary.columns), last_row])

        shorter_summary = pd.concat(
            data,
            ignore_index=True,
        )
        shorter_summary.index = cast(Index, indexes)

        return shorter_summary



import warnings
from collections.abc import Callable
from typing import Any, cast
from urllib.parse import quote

import rdflib
from rdflib import RDF, XSD, Graph, Namespace, URIRef
from rdflib.query import ResultRow

from cognite.neat._constants import UNKNOWN_TYPE
from cognite.neat._graph.queries import Queries
from cognite.neat._issues.warnings import PropertyDataTypeConversionWarning
from cognite.neat._utils.auxiliary import string_to_ideal_type
from cognite.neat._utils.rdf_ import get_namespace, remove_namespace_from_uri

from ._base import BaseTransformer, BaseTransformerStandardised, RowTransformationOutput


# TODO: Standardise
class SplitMultiValueProperty(BaseTransformer):
    description: str = (
        "SplitMultiValueProperty is a transformer that splits a "
        "multi-value property into multiple single-value properties."
    )
    _use_only_once: bool = True
    _need_changes = frozenset({})

    _object_property_template: str = """SELECT ?s ?o WHERE{{

                                ?s a <{subject_uri}> .
                                ?s <{property_uri}> ?o .
                                ?o a <{object_uri}> .

                            }}"""

    _datatype_property_template: str = """SELECT ?s ?o WHERE {{

                                ?s a <{subject_uri}> .
                                ?s <{property_uri}> ?o .
                                FILTER (datatype(?o) = <{object_uri}>)

                                }}"""

    _unknown_property_template: str = """SELECT ?s ?o WHERE {{

                                ?s a <{subject_uri}> .
                                ?s <{property_uri}> ?o .
                                FILTER NOT EXISTS {{ ?o a ?objectType }}
                                }}"""

    def transform(self, graph: Graph) -> None:
        # handle multi value type object properties
        for subject_uri, property_uri, value_types in Queries(graph).multi_value_type_property():
            for value_type_uri in value_types:
                _args = {
                    "subject_uri": subject_uri,
                    "property_uri": property_uri,
                    "object_uri": value_type_uri,
                }

                # Case 1: Unknown value type
                if value_type_uri == UNKNOWN_TYPE:
                    iterator = graph.query(self._unknown_property_template.format(**_args))

                # Case 2: Datatype value type
                elif value_type_uri.startswith(str(XSD)):
                    iterator = graph.query(self._datatype_property_template.format(**_args))

                # Case 3: Object value type
                else:
                    iterator = graph.query(self._object_property_template.format(**_args))

                for s, o in iterator:  # type: ignore [misc]
                    graph.remove((s, property_uri, o))
                    new_property = URIRef(f"{property_uri}_{remove_namespace_from_uri(value_type_uri)}")
                    graph.add((s, new_property, o))


class ConvertLiteral(BaseTransformerStandardised):
    description: str = "ConvertLiteral is a transformer that improve data typing of a literal value."
    _use_only_once: bool = False
    _need_changes = frozenset({})

    def __init__(
        self,
        subject_type: URIRef,
        subject_predicate: URIRef,
        conversion: Callable[[Any], Any] | None = None,
    ) -> None:
        self.subject_type = subject_type
        self.subject_predicate = subject_predicate
        self.conversion = conversion or string_to_ideal_type
        self._type_name = remove_namespace_from_uri(subject_type)
        self._property_name = remove_namespace_from_uri(subject_predicate)

    def _skip_count_query(self) -> str:
        query = """SELECT (COUNT(?value) AS ?valueCount)
                    WHERE {{
                      ?instance a <{subject_type}> .
                      ?instance <{subject_predicate}> ?value
                       FILTER(isIRI(?value))
                    }}"""
        return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _count_query(self) -> str:
        query = """SELECT (COUNT(?value) AS ?valueCount)
                    WHERE {{
                      ?instance a <{subject_type}> .
                      ?instance <{subject_predicate}> ?value
                       FILTER(isLiteral(?value))
                    }}"""
        return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _iterate_query(self) -> str:
        query = """SELECT ?instance ?value
                    WHERE {{
                      ?instance a <{subject_type}> .
                      ?instance <{subject_predicate}> ?value
                      FILTER(isLiteral(?value))
                    }}"""
        return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()

        instance, literal = query_result_row
        value = cast(rdflib.Literal, literal).toPython()

        try:
            converted_value = self.conversion(value)
        except Exception as e:
            warnings.warn(
                PropertyDataTypeConversionWarning(str(instance), self._type_name, self._property_name, str(e)),
                stacklevel=2,
            )
        row_output.add_triples.append((instance, self.subject_predicate, rdflib.Literal(converted_value)))  # type: ignore[arg-type]
        row_output.remove_triples.append((instance, self.subject_predicate, literal))  # type: ignore[arg-type]
        row_output.instances_modified_count += 1

        return row_output


class LiteralToEntity(BaseTransformerStandardised):
    description = "Converts a literal value to new entity"

    def __init__(
        self, subject_type: URIRef | None, subject_predicate: URIRef, entity_type: str, new_property: str | None = None
    ) -> None:
        self.subject_type = subject_type
        self.subject_predicate = subject_predicate
        self.entity_type = entity_type
        self.new_property = new_property

    def _iterate_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT ?instance ?property
            WHERE {{
              ?instance <{subject_predicate}> ?property
              FILTER(isLiteral(?property))
            }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT ?instance ?property
                WHERE {{
                  ?instance a <{subject_type}> .
                  ?instance <{subject_predicate}> ?property
                  FILTER(isLiteral(?property))
                }}"""
            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _skip_count_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT (COUNT(?property) AS ?propertyCount)
                        WHERE {{
                          ?instance <{subject_predicate}> ?property
                          FILTER(isIRI(?property))
                        }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT (COUNT(?property) AS ?propertyCount)
                        WHERE {{
                          ?instance a <{subject_type}> .
                          ?instance <{subject_predicate}> ?property
                          FILTER(isIRI(?property))
                        }}"""
            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _count_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT (COUNT(?property) AS ?propertyCount)
                WHERE {{
                  ?instance <{subject_predicate}> ?property
                  FILTER(isLiteral(?property))
                }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT (COUNT(?property) AS ?propertyCount)
                        WHERE {{
                          ?instance a <{subject_type}> .
                          ?instance <{subject_predicate}> ?property
                          FILTER(isLiteral(?property))
                        }}"""

            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()

        instance, literal = query_result_row
        value = cast(rdflib.Literal, literal).toPython()
        namespace = Namespace(get_namespace(instance))  # type: ignore[arg-type]
        entity_type = namespace[self.entity_type]
        new_entity = namespace[f"{self.entity_type}_{quote(value)!s}"]
        row_output.add_triples.append((new_entity, RDF.type, entity_type))
        row_output.instances_added_count += 1  # we add one new entity

        if self.new_property is not None:
            row_output.add_triples.append((new_entity, namespace[self.new_property], rdflib.Literal(value)))  # type: ignore[arg-type]
            row_output.instances_modified_count += 1  # we modify the new entity

        row_output.add_triples.append((instance, self.subject_predicate, new_entity))  # type: ignore[arg-type]
        row_output.remove_triples.append((instance, self.subject_predicate, literal))  # type: ignore[arg-type]
        row_output.instances_modified_count += 1  # we modify the old entity

        return row_output


class ConnectionToLiteral(BaseTransformerStandardised):
    description = "Converts an entity connection to a literal value"

    def __init__(self, subject_type: URIRef | None, subject_predicate: URIRef) -> None:
        self.subject_type = subject_type
        self.subject_predicate = subject_predicate

    def _iterate_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT ?instance ?object
            WHERE {{
              ?instance <{subject_predicate}> ?object
              FILTER(isIRI(?object))
            }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT ?instance ?object
                WHERE {{
                  ?instance a <{subject_type}> .
                  ?instance <{subject_predicate}> ?object
                  FILTER(isIRI(?object))
                }}"""
            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _skip_count_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT (COUNT(?object) AS ?objectCount)
                        WHERE {{
                          ?instance <{subject_predicate}> ?object
                          FILTER(isLiteral(?object))
                        }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT (COUNT(?object) AS ?objectCount)
                        WHERE {{
                          ?instance a <{subject_type}> .
                          ?instance <{subject_predicate}> ?object
                          FILTER(isLiteral(?object))
                        }}"""
            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _count_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT (COUNT(?object) AS ?objectCount)
                WHERE {{
                  ?instance <{subject_predicate}> ?object
                  FILTER(isIRI(?object))
                }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT (COUNT(?object) AS ?objectCount)
                        WHERE {{
                          ?instance a <{subject_type}> .
                          ?instance <{subject_predicate}> ?object
                          FILTER(isIRI(?object))
                        }}"""

            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()

        instance, object_entity = cast(tuple[URIRef, URIRef], query_result_row)
        value = remove_namespace_from_uri(object_entity)

        row_output.add_triples.append((instance, self.subject_predicate, rdflib.Literal(value)))
        row_output.remove_triples.append((instance, self.subject_predicate, object_entity))
        row_output.instances_modified_count += 1

        return row_output



from collections import Counter, defaultdict
from collections.abc import Mapping
from datetime import datetime, timezone
from pathlib import Path
from typing import ClassVar, cast

from cognite.client import data_modeling as dm
from rdflib import RDF, Namespace, URIRef
from rdflib import Literal as RdfLiteral

from cognite.neat._issues.warnings import PropertySkippedWarning, PropertyValueTypeUndefinedWarning
from cognite.neat._rules.models import data_types
from cognite.neat._rules.models.data_types import AnyURI
from cognite.neat._rules.models.entities._single_value import UnknownEntity
from cognite.neat._rules.models.information import (
    InformationMetadata,
)
from cognite.neat._store import NeatGraphStore
from cognite.neat._store._provenance import INSTANCES_ENTITY
from cognite.neat._utils.rdf_ import remove_namespace_from_uri, uri_to_short_form

from ._base import DEFAULT_NON_EXISTING_NODE_TYPE, BaseRDFImporter

DEFAULT_INFERENCE_DATA_MODEL_ID = ("neat_space", "InferredDataModel", "inferred")

ORDERED_CLASSES_QUERY = """SELECT ?class (count(?s) as ?instances )
                           WHERE { ?s a ?class . }
                           group by ?class order by DESC(?instances)"""


INSTANCES_OF_CLASS_QUERY = """SELECT ?s ?propertyCount WHERE { ?s a <class> . BIND ('Unknown' as ?propertyCount) }"""


INSTANCES_OF_CLASS_RICHNESS_ORDERED_QUERY = """SELECT ?s (COUNT(?p) as ?propertyCount)
                                               WHERE { ?s a <class> ; ?p ?o . }
                                               GROUP BY ?s
                                               ORDER BY DESC(?propertyCount)"""

INSTANCE_PROPERTIES_DEFINITION = """SELECT ?property (count(?property) as ?occurrence) ?dataType ?objectType
                                    WHERE {<instance_id> ?property ?value .

                                           BIND(datatype(?value) AS ?dataType)

                                           OPTIONAL {?value rdf:type ?objectType .}}
                                    GROUP BY ?property ?dataType ?objectType"""


class InferenceImporter(BaseRDFImporter):
    """Infers rules from a triple store.

    Rules inference through analysis of knowledge graph provided in various formats.
    Use the factory methods to create a triple store from sources such as
    RDF files, JSON files, YAML files, XML files, or directly from a graph store.

    ClassVars:
        overwrite_data_types: Mapping of data types to be overwritten. The InferenceImporter will overwrite
            32-bit integer and 32-bit float data types to 64-bit integer and 64-bit float data types

    Args:
        issue_list: Issue list to store issues
        graph: Knowledge graph
        max_number_of_instance: Maximum number of instances to be used in inference
        prefix: Prefix to be used for the inferred model


    """

    overwrite_data_types: ClassVar[Mapping[URIRef, URIRef]] = {
        data_types.Integer.as_xml_uri_ref(): data_types.Long.as_xml_uri_ref(),
        data_types.Float.as_xml_uri_ref(): data_types.Double.as_xml_uri_ref(),
    }

    @classmethod
    def from_graph_store(
        cls,
        store: NeatGraphStore,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_INFERENCE_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
        language: str = "en",
    ) -> "InferenceImporter":
        return super().from_graph_store(
            store,
            data_model_id,
            max_number_of_instance,
            non_existing_node_type,
            language,
        )

    @classmethod
    def from_file(
        cls,
        filepath: Path,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_INFERENCE_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
        language: str = "en",
        source_name: str = "Unknown",
    ) -> "InferenceImporter":
        return super().from_file(
            filepath,
            data_model_id,
            max_number_of_instance,
            non_existing_node_type,
            language,
            source_name=source_name,
        )

    @classmethod
    def from_json_file(
        cls,
        filepath: Path,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_INFERENCE_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
        language: str = "en",
    ) -> "InferenceImporter":
        raise NotImplementedError("JSON file format is not supported yet.")

    @classmethod
    def from_yaml_file(
        cls,
        filepath: Path,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_INFERENCE_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
        language: str = "en",
    ) -> "InferenceImporter":
        raise NotImplementedError("YAML file format is not supported yet.")

    @classmethod
    def from_xml_file(
        cls,
        filepath: Path,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_INFERENCE_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
    ) -> "InferenceImporter":
        raise NotImplementedError("JSON file format is not supported yet.")

    def _to_rules_components(
        self,
    ) -> dict:
        """Convert RDF graph to dictionary defining data model and prefixes of the graph

        Args:
            graph: RDF graph to be converted to TransformationRules object
            max_number_of_instance: Max number of instances to be considered for each class

        Returns:
            Tuple of data model and prefixes of the graph
        """

        classes: dict[str, dict] = {}
        properties: dict[str, dict] = {}
        prefixes: dict[str, Namespace] = {}
        count_by_value_type_by_property: dict[str, dict[str, int]] = defaultdict(Counter)

        # Infers all the classes in the graph
        for class_uri, no_instances in self.graph.query(ORDERED_CLASSES_QUERY):  # type: ignore[misc]
            if (class_id := remove_namespace_from_uri(cast(URIRef, class_uri))) in classes:
                # handles cases when class id is already present in classes
                class_id = f"{class_id}_{len(classes)+1}"

            classes[class_id] = {
                "class_": class_id,
                "uri": class_uri,
                "comment": f"Inferred from knowledge graph, where this class has <{no_instances}> instances",
            }

            self._add_uri_namespace_to_prefixes(cast(URIRef, class_uri), prefixes)

        instances_query = (
            INSTANCES_OF_CLASS_QUERY if self.max_number_of_instance == -1 else INSTANCES_OF_CLASS_RICHNESS_ORDERED_QUERY
        )

        # Infers all the properties of the class
        for class_id, class_definition in classes.items():
            for (  # type: ignore[misc]
                instance,
                _,
            ) in self.graph.query(  # type: ignore[misc]
                instances_query.replace("class", class_definition["uri"])
                if self.max_number_of_instance < 0
                else instances_query.replace("class", class_definition["uri"]) + f" LIMIT {self.max_number_of_instance}"
            ):
                for property_uri, occurrence, data_type_uri, object_type_uri in self.graph.query(  # type: ignore[misc]
                    INSTANCE_PROPERTIES_DEFINITION.replace("instance_id", instance)
                ):  # type: ignore[misc]
                    # this is to skip rdf:type property
                    if property_uri == RDF.type:
                        continue
                    property_id = remove_namespace_from_uri(property_uri)
                    if property_id in {"external_id", "externalId"}:
                        skip_issue = PropertySkippedWarning(
                            resource_type="Property",
                            identifier=f"{class_id}:{property_id}",
                            property_name=property_id,
                            reason="External ID is assumed to be the unique identifier of the instance "
                            "and is not part of the data model schema.",
                        )
                        if skip_issue not in self.issue_list:
                            self.issue_list.append(skip_issue)
                        continue

                    self._add_uri_namespace_to_prefixes(cast(URIRef, property_uri), prefixes)

                    if isinstance(data_type_uri, URIRef):
                        data_type_uri = self.overwrite_data_types.get(data_type_uri, data_type_uri)

                    if value_type_uri := (data_type_uri or object_type_uri):
                        self._add_uri_namespace_to_prefixes(cast(URIRef, value_type_uri), prefixes)
                        value_type_id = remove_namespace_from_uri(value_type_uri)

                    # this handles situations when property points to node that is not present in graph
                    else:
                        value_type_id = str(self.non_existing_node_type)

                        issue = PropertyValueTypeUndefinedWarning(
                            resource_type="Property",
                            identifier=f"{class_id}:{property_id}",
                            property_name=property_id,
                            default_action="Remove the property from the rules",
                            recommended_action="Make sure that graph is complete",
                        )

                        if issue not in self.issue_list:
                            self.issue_list.append(issue)

                    id_ = f"{class_id}:{property_id}"

                    definition = {
                        "class_": class_id,
                        "property_": property_id,
                        "max_count": cast(RdfLiteral, occurrence).value,
                        "value_type": value_type_id,
                        "transformation": (
                            f"{uri_to_short_form(class_definition['uri'], prefixes)}"
                            f"({uri_to_short_form(cast(URIRef, property_uri), prefixes)})"
                        ),
                    }

                    count_by_value_type_by_property[id_][value_type_id] += 1

                    # USE CASE 1: If property is not present in properties
                    if id_ not in properties:
                        definition["value_type"] = {definition["value_type"]}
                        properties[id_] = definition

                    # USE CASE 2: first time redefinition, value type change to multi
                    elif id_ in properties and definition["value_type"] not in properties[id_]["value_type"]:
                        properties[id_]["value_type"].add(definition["value_type"])

                    # USE CASE 3: existing but max count is different
                    elif (
                        id_ in properties
                        and definition["value_type"] in properties[id_]["value_type"]
                        and properties[id_]["max_count"] != definition["max_count"]
                    ):
                        properties[id_]["max_count"] = max(properties[id_]["max_count"], definition["max_count"])

        # Create multi-value properties otherwise single value
        for property_ in properties.values():
            # Removes non-existing node type from value type prior final conversion to string
            if len(property_["value_type"]) > 1 and str(self.non_existing_node_type) in property_["value_type"]:
                property_["value_type"].remove(str(self.non_existing_node_type))

            if len(property_["value_type"]) > 1:
                property_["value_type"] = " | ".join([str(t) for t in property_["value_type"]])
            else:
                property_["value_type"] = next(iter(property_["value_type"]))

        return {
            "metadata": self._default_metadata().model_dump(),
            "classes": list(classes.values()),
            "properties": list(properties.values()),
            "prefixes": prefixes,
        }

    def _default_metadata(self):
        now = datetime.now(timezone.utc)
        return InformationMetadata(
            space=self.data_model_id.space,
            external_id=self.data_model_id.external_id,
            version=self.data_model_id.version,
            name="Inferred Model",
            creator="NEAT",
            created=now,
            updated=now,
            description="Inferred model from knowledge graph",
        )

    @property
    def source_uri(self) -> URIRef:
        return INSTANCES_ENTITY.id_



from collections.abc import Callable, Collection
from typing import Any, Literal, cast

from cognite.client.data_classes.data_modeling import DataModelIdentifier
from rdflib import URIRef

from cognite.neat._client import NeatClient
from cognite.neat._constants import (
    get_default_prefixes_and_namespaces,
)
from cognite.neat._graph import extractors
from cognite.neat._graph.transformers import (
    AttachPropertyFromTargetToSource,
    ConnectionToLiteral,
    ConvertLiteral,
    LiteralToEntity,
    PruneDeadEndEdges,
    PruneInstancesOfUnknownType,
    PruneTypes,
    RelationshipAsEdgeTransformer,
    Transformers,
)
from cognite.neat._graph.transformers._rdfpath import MakeConnectionOnExactMatch
from cognite.neat._issues import IssueList
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._rules.models.dms import DMSValidation
from cognite.neat._rules.transformers import (
    AddClassImplements,
    IncludeReferenced,
    PrefixEntities,
    ReduceCogniteModel,
    RulesTransformer,
    ToCompliantEntities,
    ToDataProductModel,
    ToEnterpriseModel,
    ToSolutionModel,
)
from cognite.neat._utils.text import humanize_collection

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class PrepareAPI:
    """Apply various operations on the knowledge graph as a necessary preprocessing step before for instance
    inferring a data model or exporting the knowledge graph to a desired destination.
    """

    def __init__(self, client: NeatClient | None, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self.data_model = DataModelPrepareAPI(client, state, verbose)
        self.instances = InstancePrepareAPI(state, verbose)


@session_class_wrapper
class InstancePrepareAPI:
    """Operations to perform on instances of data in the knowledge graph."""

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose

    def dexpi(self) -> None:
        """Prepares extracted DEXPI graph for further usage in CDF

        !!! note "This method bundles several graph transformers which"
            - attach values of generic attributes to nodes
            - create associations between nodes
            - remove unused generic attributes
            - remove associations between nodes that do not exist in the extracted graph
            - remove edges to nodes that do not exist in the extracted graph

        and therefore safeguard CDF from a bad graph

        Example:
            Apply Dexpi specific transformations:
            ```python
            neat.prepare.instances.dexpi()
            ```
        """

        DEXPI = get_default_prefixes_and_namespaces()["dexpi"]

        transformers = [
            # Remove any instance which type is unknown
            PruneInstancesOfUnknownType(),
            # Directly connect generic attributes
            AttachPropertyFromTargetToSource(
                target_property=DEXPI.Value,
                target_property_holding_new_property=DEXPI.Name,
                target_node_type=DEXPI.GenericAttribute,
                delete_target_node=True,
            ),
            # Directly connect associations
            AttachPropertyFromTargetToSource(
                target_property=DEXPI.ItemID,
                target_property_holding_new_property=DEXPI.Type,
                target_node_type=DEXPI.Association,
                delete_target_node=True,
            ),
            # Remove unused generic attributes and associations
            PruneTypes([DEXPI.GenericAttribute, DEXPI.Association]),
            # Remove edges to nodes that do not exist in the extracted graph
            PruneDeadEndEdges(),
        ]

        for transformer in transformers:
            self._state.instances.store.transform(cast(Transformers, transformer))

    def aml(self) -> None:
        """Prepares extracted AutomationML graph for further usage in CDF

        !!! note "This method bundles several graph transformers which"
            - attach values of attributes to nodes
            - remove unused attributes
            - remove edges to nodes that do not exist in the extracted graph

        and therefore safeguard CDF from a bad graph

        Example:
            Apply AML specific transformations:
            ```python
            neat.prepare.instances.aml()
            ```
        """

        AML = get_default_prefixes_and_namespaces()["aml"]

        transformers = [
            # Remove any instance which type is unknown
            PruneInstancesOfUnknownType(),
            # Directly connect generic attributes
            AttachPropertyFromTargetToSource(
                target_property=AML.Value,
                target_property_holding_new_property=AML.Name,
                target_node_type=AML.Attribute,
                delete_target_node=True,
            ),
            # Prune unused attributes
            PruneTypes([AML.Attribute]),
            # # Remove edges to nodes that do not exist in the extracted graph
            PruneDeadEndEdges(),
        ]

        for transformer in transformers:
            self._state.instances.store.transform(cast(Transformers, transformer))

    def make_connection_on_exact_match(
        self,
        source: tuple[str, str],
        target: tuple[str, str],
        connection: str | None = None,
        limit: int | None = 100,
    ) -> None:
        """Make connection on exact match.

        Args:
            source: The source of the connection. A tuple of (type, property) where
                    where property is the property that should be matched on the source
                    to make the connection with the target.
            target: The target of the connection. A tuple of (type, property) where
                    where property is the property that should be matched on the target
                    to make the connection with the source.

            connection: new property to use for the connection. If None, the connection
                    will be made by lowercasing the target type.
            limit: The maximum number of connections to make. If None, all connections

        !!! note "Make Connection on Exact Match"
            This method will make a connection between the source and target based on the exact match:
            (SourceType)-[sourceProperty]->(sourceValue) == (TargetType)-[targetProperty]->(targetValue)

            The connection will be made by creating a new property on the source type that will contain the
            target value, as follows:
            (SourceType)-[connection]->(TargetType)

        Example:
            Make connection on exact match:
            ```python
            # From an active NeatSession
            neat.read.csv("workitem.Table.csv",
                          type = "Activity",
                          primary_key="sourceId")

            neat.read.csv("assets.Table.csv",
                          type="Asset",
                          primary_key="WMT_TAG_GLOBALID")

            # Here we specify what column from the source table we should use when we link it with a column in the
            # target table. In this case, it is the "workorderItemname" column in the source table
            source = ("Activity", "workorderItemname")

            # Here we give a name to the new property that is created when a match between the source and target is
            # found
            connection = "asset"

            # Here we specify what column from the target table we should use when searching for a match.
            # In this case, it is the "wmtTagName" column in the target table
            target = ("Asset", "wmtTagName")

            neat.prepare.instances.make_connection_on_exact_match(source, target, connection)
            ```
        """
        try:
            subject_type, subject_predicate = self._get_type_and_property_uris(*source)
            object_type, object_predicate = self._get_type_and_property_uris(*target)
        except NeatValueError as e:
            raise NeatSessionError(f"Cannot make connection: {e}") from None

        transformer = MakeConnectionOnExactMatch(
            subject_type,
            subject_predicate,
            object_type,
            object_predicate,
            connection,
            limit,
        )

        self._state.instances.store.transform(transformer)

    def _get_type_and_property_uris(self, type_: str, property_: str) -> tuple[URIRef, URIRef]:
        type_uri = self._state.instances.store.queries.type_uri(type_)
        property_uri = self._state.instances.store.queries.property_uri(property_)

        if not type_uri:
            raise NeatValueError(f"Type {type_} does not exist in the graph.")
        elif len(type_uri) > 1:
            raise NeatValueError(f"{type_} has multiple ids found in the graph: {humanize_collection(type_uri)}.")

        if not property_uri:
            raise NeatValueError(f"Property {property_} does not exist in the graph.")
        elif len(type_uri) > 1:
            raise NeatValueError(
                f"{property_} has multiple ids found in the graph: {humanize_collection(property_uri)}."
            )

        if not self._state.instances.store.queries.type_with_property(type_uri[0], property_uri[0]):
            raise NeatValueError(f"Property {property_} is not defined for type {type_}.")
        return type_uri[0], property_uri[0]

    def relationships_as_edges(self, min_relationship_types: int = 1, limit_per_type: int | None = None) -> None:
        """This assumes that you have read a classic CDF knowledge graph including relationships.

        This method converts relationships into edges in the graph. This is useful as the
        edges will be picked up as part of the schema connected to Assets, Events, Files, Sequences,
        and TimeSeries in the InferenceImporter.

        Args:
            min_relationship_types: The minimum number of relationship types that must exists to convert those
                relationships to edges. For example, if there is only 5 relationships between Assets and TimeSeries,
                and limit is 10, those relationships will not be converted to edges.
            limit_per_type: The number of conversions to perform per relationship type. For example, if there are 10
                relationships between Assets and TimeSeries, and limit_per_type is 1, only 1 of those relationships
                will be converted to an edge. If None, all relationships will be converted.

        """
        transformer = RelationshipAsEdgeTransformer(min_relationship_types, limit_per_type)
        self._state.instances.store.transform(transformer)

    def convert_data_type(self, source: tuple[str, str], *, convert: Callable[[Any], Any] | None = None) -> None:
        """Convert the data type of the given property.

        This is, for example, useful when you have a boolean property that you want to convert to an enum.

        Args:
            source: The source of the conversion. A tuple of (type, property)
                    where property is the property that should be converted.
            convert: The function to use for the conversion. The function should take the value of the property
                    as input and return the converted value. Default to assume you have a string that should be
                    converted to int, float, bool, or datetime.

        Example:
            Convert a boolean property to a string:
            ```python
            neat.prepare.instances.convert_data_type(
                ("TimeSeries", "isString"),
                convert=lambda is_string: "string" if is_string else "numeric"
            )
            ```

        """
        try:
            subject_type, subject_predicate = self._get_type_and_property_uris(*source)
        except NeatValueError as e:
            raise NeatSessionError(f"Cannot convert data type: {e}") from None

        transformer = ConvertLiteral(subject_type, subject_predicate, convert)
        self._state.instances.store.transform(transformer)

    def property_to_type(self, source: tuple[str | None, str], type: str, new_property: str | None = None) -> None:
        """Convert a property to a new type.

        Args:
            source: The source of the conversion. A tuple of (type, property)
                    where property is the property that should be converted.
                    You can pass (None, property) to covert all properties with the given name.
            type: The new type of the property.
            new_property: Add the identifier as a new property. If None, the new entity will not have a property.

        Example:
            Convert the property 'source' to SourceSystem
            ```python
            neat.prepare.instances.property_to_type(
                (None, "source"), "SourceSystem"
            )
            ```
        """
        subject_type: URIRef | None = None
        if source[0] is not None:
            try:
                subject_type, subject_predicate = self._get_type_and_property_uris(*source)  # type: ignore[arg-type, assignment]
            except NeatValueError as e:
                raise NeatSessionError(f"Cannot convert to type: {e}") from None
        else:
            subject_predicate = self._state.instances.store.queries.property_uri(source[1])[0]

        transformer = LiteralToEntity(subject_type, subject_predicate, type, new_property)
        self._state.instances.store.transform(transformer)

    def connection_to_data_type(self, source: tuple[str | None, str]) -> None:
        """Converts a connection to a data type.

        Args:
            source: The source of the conversion. A tuple of (type, property)
                    where property is the property that should be converted.
                    You can pass (None, property) to covert all properties with the given name.

        Example:

            Convert all properties 'labels' from a connection to a string:

            ```python
            neat.prepare.instances.connection_to_data_type(
                (None, "labels")
            )
            ```

        """
        subject_type: URIRef | None = None
        if source[0] is not None:
            try:
                subject_type, subject_predicate = self._get_type_and_property_uris(*source)  # type: ignore[arg-type, assignment]
            except NeatValueError as e:
                raise NeatSessionError(f"Cannot convert to data type: {e}") from None
        else:
            subject_predicate = self._state.instances.store.queries.property_uri(source[1])[0]
        transformer = ConnectionToLiteral(subject_type, subject_predicate)
        self._state.instances.store.transform(transformer)

    def classic_to_core(self) -> None:
        """Prepares extracted CDF classic graph for the Core Data model.

        !!! note "This method bundles several graph transformers which"
            - Convert relationships to edges
            - Convert TimeSeries.type from bool to enum
            - Convert all properties 'source' to a connection to SourceSystem
            - Convert all properties 'labels' from a connection to a string

        Example:
            Apply classic to core transformations:
            ```python
            neat.prepare.instances.classic_to_core()
            ```
        """
        self.relationships_as_edges()
        self.convert_data_type(
            ("TimeSeries", "isString"), convert=lambda is_string: "string" if is_string else "numeric"
        )
        self.property_to_type((None, "source"), "SourceSystem", "name")
        for type_ in [
            extractors.EventsExtractor._default_rdf_type,
            extractors.AssetsExtractor._default_rdf_type,
            extractors.FilesExtractor._default_rdf_type,
        ]:
            try:
                subject_type, subject_predicate = self._get_type_and_property_uris(type_, "labels")
            except NeatValueError:
                # If the type_.labels does not exist, continue. This is not an error, it just means that the
                # Labels is not used in the graph for that type.
                continue
            else:
                transformer = ConnectionToLiteral(subject_type, subject_predicate)
                self._state.instances.store.transform(transformer)


@session_class_wrapper
class DataModelPrepareAPI:
    """Operations to perform on a data model as part of a workflow before writing the data model
    to a desired destination.
    """

    def __init__(self, client: NeatClient | None, state: SessionState, verbose: bool) -> None:
        self._client = client
        self._state = state
        self._verbose = verbose

    def cdf_compliant_external_ids(self) -> IssueList:
        """Convert data model component external ids to CDF compliant entities."""
        return self._state.rule_transform(ToCompliantEntities())

    def prefix(self, prefix: str) -> IssueList:
        """Prefix all views in the data model with the given prefix.

        Args:
            prefix: The prefix to add to the views in the data model.

        """
        return self._state.rule_transform(PrefixEntities(prefix))

    def to_enterprise(
        self,
        data_model_id: DataModelIdentifier,
        org_name: str = "My",
        dummy_property: str = "GUID",
        move_connections: bool = False,
    ) -> IssueList:
        """Uses the current data model as a basis to create enterprise data model

        Args:
            data_model_id: The enterprise data model id that is being created
            org_name: Organization name to use for the views in the enterprise data model.
            dummy_property: The dummy property to use as placeholder for the views in the new data model.
            move_connections: If True, the connections will be moved to the new data model.

        !!! note "Enterprise Data Model Creation"

            Always create an enterprise data model from a Cognite Data Model as this will
            assure all the Cognite Data Fusion applications to run smoothly, such as
                - Search
                - Atlas AI
                - ...

        !!! note "Move Connections"

            If you want to move the connections to the new data model, set the move_connections
            to True. This will move the connections to the new data model and use new model
            views as the source and target views.

        """
        return self._state.rule_transform(
            ToEnterpriseModel(
                new_model_id=data_model_id,
                org_name=org_name,
                dummy_property=dummy_property,
                move_connections=move_connections,
            )
        )

    def to_solution(
        self,
        data_model_id: DataModelIdentifier,
        org_name: str = "My",
        mode: Literal["read", "write"] = "read",
        dummy_property: str = "GUID",
    ) -> IssueList:
        """Uses the current data model as a basis to create solution data model

        Args:
            data_model_id: The solution data model id that is being created.
            org_name: Organization name to use for the views in the new data model.
            mode: The mode of the solution data model. Can be either "read" or "write".
            dummy_property: The dummy property to use as placeholder for the views in the new data model.

        !!! note "Solution Data Model Mode"

            The read-only solution model will only be able to read from the existing containers
            from the enterprise data model, therefore the solution data model will not have
            containers in the solution data model space. Meaning the solution data model views
            will be read-only.

            The write mode will have additional containers in the solution data model space,
            allowing in addition to reading through the solution model views, also writing to
            the containers in the solution data model space.

        """
        return self._state.rule_transform(
            ToSolutionModel(
                new_model_id=data_model_id,
                org_name=org_name,
                mode=mode,
                dummy_property=dummy_property,
            )
        )

    def to_data_product(
        self,
        data_model_id: DataModelIdentifier,
        org_name: str = "",
        include: Literal["same-space", "all"] = "same-space",
    ) -> None:
        """Uses the current data model as a basis to create data product data model.

        A data product model is a data model that ONLY maps to containers and do not use implements. This is
        typically used for defining the data in a data product.

        Args:
            data_model_id: The data product data model id that is being created.
            org_name: Organization name used as prefix if the model is building on top of a Cognite Data Model.
            include: The views to include in the data product data model. Can be either "same-space" or "all".
                If you set same-space, only the properties of the views in the same space as the data model
                will be included.
        """

        view_ids, container_ids = DMSValidation(
            self._state.rule_store.last_verified_dms_rules
        ).imported_views_and_containers_ids()
        transformers: list[RulesTransformer] = []
        if (view_ids or container_ids) and self._client is None:
            raise NeatSessionError(
                "No client provided. You are referencing unknown views and containers in your data model, "
                "NEAT needs a client to lookup the definitions. "
                "Please set the client in the session, NeatSession(client=client)."
            )
        elif (view_ids or container_ids) and self._client:
            transformers.append(IncludeReferenced(self._client, include_properties=True))

        transformers.append(
            ToDataProductModel(
                new_model_id=data_model_id,
                org_name=org_name,
                include=include,
            )
        )

        self._state.rule_transform(*transformers)

    def reduce(self, drop: Collection[Literal["3D", "Annotation", "BaseViews"] | str]) -> IssueList:
        """This is a special method that allow you to drop parts of the data model.
        This only applies to Cognite Data Models.

        Args:
            drop: What to drop from the data model. The values 3D, Annotation, and BaseViews are special values that
                drops multiple views at once. You can also pass externalIds of views to drop individual views.

        """
        return self._state.rule_transform(ReduceCogniteModel(drop))

    def include_referenced(self) -> IssueList:
        """Include referenced views and containers in the data model."""
        if self._client is None:
            raise NeatSessionError(
                "No client provided. You are referencing unknown views and containers in your data model, "
                "NEAT needs a client to lookup the definitions. "
                "Please set the client in the session, NeatSession(client=client)."
            )
        return self._state.rule_transform(IncludeReferenced(self._client))

    def add_implements_to_classes(self, suffix: Literal["Edge"], implements: str = "Edge") -> IssueList:
        """All classes with the suffix will have the implements property set to the given value.

        Args:
            suffix: The suffix of the classes to add the implements property to.
            implements:  The value of the implements property to set.

        """
        return self._state.rule_transform(AddClassImplements(implements, suffix))
